#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
import six
import os
import io
import codecs
from inspect import getsource

# for custom indexes
from bitdust_forks.CodernityDB3.storage import Storage, IU_Storage
from bitdust_forks.CodernityDB3.hash_index import (
    HashIndex,
    IU_UniqueHashIndex,
    IU_HashIndex,
    UniqueHashIndex)
# normal imports

from bitdust_forks.CodernityDB3.index import (
    ElemNotFound,
    DocIdNotFound,
    IndexException,
    Index,
    TryReindexException,
    ReindexException,
    IndexNotFoundException,
    IndexConflict)

from bitdust_forks.CodernityDB3.misc import NONE

from bitdust_forks.CodernityDB3.env import cdb_environment

from random import randrange

import warnings
import collections
import codecs


def header_for_indexes(index_name, index_class, db_custom="", ind_custom="", classes_code=""):
    return """# %s
# %s

# inserted automatically
import os
import marshal

import struct
import shutil

from hashlib import md5

# custom db code start
# db_custom
%s

# custom index code start
# ind_custom
%s

# source of classes in index.classes_code
# classes_code
%s

# index code start

""" % (index_name, index_class, db_custom, ind_custom, classes_code)


class DatabaseException(Exception):
    pass


class PreconditionsException(DatabaseException):
    pass


class RecordDeleted(DatabaseException):
    pass


class RecordNotFound(DatabaseException):
    pass


class RevConflict(DatabaseException):
    pass


class DatabaseConflict(DatabaseException):
    pass


class DatabasePathException(DatabaseException):
    pass


class DatabaseIsNotOpened(PreconditionsException):
    pass


class Database(object):
    """
    A default single thread database object.
    """

    custom_header = ""  # : use it for imports required by your database

    def __init__(self, path):
        self.path = path
        self.storage = None
        self.indexes = []
        self.id_ind = None
        self.indexes_names = {}
        self.opened = False

    def create_new_rev(self, old_rev=None):
        """
        Creates new revision number based on previous one.
        Increments it + random bytes. On overflow starts from 0 again.
        """
        if old_rev:
            try:
                rev_num = int(old_rev[:4], 16)
            except:
                raise RevConflict()
            rev_num += 1
            if rev_num > 65025:
            # starting the counter from 0 again
                rev_num = 0
            rnd = randrange(65536)
            return "%04x%04x" % (rev_num, rnd)
        else:
            # new rev
            rnd = randrange(256 ** 2)
            return '0001%04x' % rnd

    def __not_opened(self):
        if not self.opened:
            raise DatabaseIsNotOpened("Database is not opened")

    def set_indexes(self, indexes=[]):
        """
        Set indexes using ``indexes`` param

        :param indexes: indexes to set in db
        :type indexes: iterable of :py:class:`CodernityDB3.index.Index` objects.

        """
        for ind in indexes:
            self.add_index(ind, create=False)

    def _add_single_index(self, p, i, index):
        """
        Adds single index to a database.
        It will use :py:meth:`inspect.getsource` to get class source.
        Then it will build real index file, save it in ``_indexes`` directory.
        """
        code = getsource(index.__class__)
        if not code.startswith('c'):  # fix for indented index codes
            import textwrap
            code = textwrap.dedent(code)
        index._order = i
        cls_code = getattr(index, 'classes_code', [])
        classes_code = ""
        for curr in cls_code:
            classes_code += getsource(curr) + '\n\n'
        with io.FileIO(os.path.join(p, "%.2d%s" % (i, index.name) + '.py'), 'w') as f:
            f.write(header_for_indexes(index.name,
                                       index.__class__.__name__,
                                       getattr(self, 'custom_header', ''),
                                       getattr(index, 'custom_header', ''),
                                       classes_code).encode())
            f.write(code.encode())
        return True

    def _read_index_single(self, p, ind, ind_kwargs={}):
        """
        It will read single index from index file (ie. generated in :py:meth:`._add_single_index`).
        Then it will perform ``exec`` on that code

        If error will occur the index file will be saved with ``_broken`` suffix

        :param p: path
        :param ind: index name (will be joined with *p*)
        :returns: new index object
        """
        with io.FileIO(os.path.join(p, ind), 'r') as f:
            name = f.readline()[2:].strip()
            _class = f.readline()[2:].strip()
            code = f.read()
        try:
            obj = compile(code, '<Index: %s' % os.path.join(p, ind), 'exec')
            exec(obj, globals())
            ind_obj = globals()[_class.decode()](self.path, name, **ind_kwargs)
            ind_obj._order = int(ind[:2])
        except:
            ind_path = os.path.join(p, ind)
            os.rename(ind_path, ind_path + '_broken')  # rename it instead of removing
#            os.unlink(os.path.join(p, ind))
            warnings.warn("Fatal error in index, saved as %s" % (ind_path + '_broken', ))
            raise
        else:
            return ind_obj

    def __check_if_index_unique(self, name, num):
        indexes = os.listdir(os.path.join(self.path, '_indexes'))
        if any((x for x in indexes if x[2:-3] == name and x[:2] != str(num))):
            raise IndexConflict("Already exists")

    def __write_index(self, new_index, number=0, edit=False, ind_kwargs=None):
        if ind_kwargs is None:
            ind_kwargs = {}
        p = os.path.join(self.path, '_indexes')
        if isinstance(new_index, str) and not new_index.startswith("path:"):
            if len(new_index.splitlines()) < 4 or new_index.splitlines()[3] != '# inserted automatically':
                from .indexcreator import Parser
                par = Parser()
                # s = par.parse(new_index).splitlines()
                custom_imports, s = par.parse(new_index)
                s = s.splitlines()
                name = s[0][2:]
                c = s[1][2:]
                comented = ['\n\n#SIMPLIFIED CODE']
                list([comented.append("#" + x) for x in new_index.splitlines()])
                comented.append('#SIMPLIFIED CODE END\n\n')

                s = header_for_indexes(
                    name, c, ind_custom=custom_imports) + "\n".join(s[2:]) + "\n".join(comented)
                new_index = s
            else:
                name = new_index.splitlines()[0][2:]
                name = name.strip()

            if name in self.indexes_names and not edit:
                raise IndexConflict("Already exists")
            if edit:
                previous_index = [x for x in os.listdir(p) if x.endswith(
                    '.py') and x[2:-3] == name]
                if not previous_index:
                    raise PreconditionsException(
                        "Can't edit index that's not yet in database")
                number = int(previous_index[0][:2])
            if number == 0 and not edit and not name == 'id':
                raise PreconditionsException(
                    "Id index must be the first added")
            ind_path = "%.2d%s" % (number, name)
            if not edit:
                self.__check_if_index_unique(name, number)

            ind_path_f = os.path.join(p, ind_path + '.py')
            if os.path.exists(ind_path_f):
                os.rename(ind_path_f, ind_path_f + '_last')  # save last working index code
            with io.FileIO(ind_path_f, 'w') as f:
                f.write(codecs.encode(new_index, 'utf-8'))

            ind_obj = self._read_index_single(p, ind_path + '.py')

        elif isinstance(new_index, str) and new_index.startswith("path:"):
            path = new_index[5:]
            if not path.endswith('.py'):
                path += '.py'
            ind_obj = self._read_index_single(p, path, ind_kwargs)
            name = ind_obj.name
            if name in self.indexes_names and not edit:
                raise IndexConflict("Already exists")
        elif isinstance(new_index, Index):
            # it will first save index as a string, and then compile it
            # it will allow to control the index object on the DB side
            ind = new_index
            init_arguments = new_index.__class__.__init__.__code__.co_varnames[
                3:]  # ignore self, path and name
            for curr in init_arguments:
                if curr not in ('args', 'kwargs'):
                    v = getattr(ind, curr, NONE())
                    if not isinstance(v, NONE):
                        ind_kwargs[curr] = v
            if edit:
                # code duplication...
                previous_index = [x for x in os.listdir(p) if x.endswith(
                    '.py') and x[2:-3] == ind.name]
                if not previous_index:
                    raise PreconditionsException(
                        "Can't edit index that's not yet in database")
                number = int(previous_index[0][:2])
            if ind.name in self.indexes_names and not edit:
                raise IndexConflict("Already exists")
            if number == 0 and not edit and not ind.name == 'id':
                raise PreconditionsException(
                    "Id index must be the first added")
            if not edit:
                self.__check_if_index_unique(ind.name, number)
            self._add_single_index(p, number, ind)
            ind_path = "%.2d%s" % (number, ind.name)
            ind_obj = self._read_index_single(p, ind_path + '.py', ind_kwargs)
            name = ind_obj.name
        else:
            raise PreconditionsException("Argument must be Index instance, path to index_file or valid string index format")
        return ind_obj, name

    def add_index(self, new_index, create=True, ind_kwargs=None):
        """

        :param new_index: New index to add, can be Index object, index valid string or path to file with index code
        :type new_index: string
        :param create: Create the index after add or not
        :type create: bool

        :returns: new index name
        """

        if ind_kwargs is None:
            ind_kwargs = {}
        p = os.path.join(self.path, '_indexes')
        if not os.path.exists(p):
            self.initialize()
        current = sorted([x for x in os.listdir(p) if x.endswith('.py')])
        if current:
            last = int(current[-1][:2])  # may crash... ignore
            _next = last + 1
        else:
            _next = 0
        ind_obj, name = self.__write_index(new_index, _next, edit=False)
        # add the new index to objects
        self.indexes.append(ind_obj)
        self.indexes_names[name] = ind_obj
        if create:
            if self.exists():  # no need te create if database doesn't exists'
                ind_obj.create_index()
        if name == 'id':
            self.__set_main_storage()
            self.__compat_things()
        for patch in getattr(ind_obj, 'patchers', ()):  # index can patch db object
            patch(self)
        return name

    def edit_index(self, index, reindex=False, ind_kwargs=None):
        """
        Allows to edit existing index.
        Previous working version will be saved with ``_last`` suffix (see :py:meth:`.revert_index`

        :param bool reindex: should be the index reindexed after change

        :returns: index name
        """
        if ind_kwargs is None:
            ind_kwargs = {}
        ind_obj, name = self.__write_index(index, -1, edit=True)
        old = next(x for x in self.indexes if x.name == name)
        old.close_index()
        index_of_index = self.indexes.index(old)
        ind_obj.open_index()
        self.indexes[index_of_index] = ind_obj
        self.indexes_names[name] = ind_obj
        if reindex:
            self.reindex_index(name)
        return name

    def revert_index(self, index_name, reindex=False, ind_kwargs=None):
        """
        Tries to revert index code from copy.
        It calls :py:meth:`.edit_index` with previous working.

        :param string index_name: index name to restore
        """
        ind_path = os.path.join(self.path, '_indexes')
        if index_name in self.indexes_names:  # then it's working index.
            ind = self.indexes_names[index_name]
            full_name = '%.2d%s.py' % (ind._order, index_name)
        else:
            indexes = os.listdir(ind_path)
            full_name = next((x for x in indexes if x[2:-3] == index_name))
        if not full_name:
            raise DatabaseException("%s index not found" % index_name)
        last_path = os.path.join(ind_path, full_name + "_last")
        if not os.path.exists(last_path):
            raise DatabaseException("No previous copy found for %s" % index_name)
        correct_last_path = last_path[:-5]  # remove _last from name
        os.rename(last_path, correct_last_path)
#        ind_data = open(last_path, 'r')
        p = 'path:%s' % os.path.split(correct_last_path)[1]
        return self.edit_index(p, reindex, ind_kwargs)

    def get_index_code(self, index_name, code_switch='All'):
        """
        It will return full index code from index file.

        :param index_name: the name of index to look for code
        """
        if not index_name in self.indexes_names:
            self.__not_opened()
            raise IndexNotFoundException(
                "Index `%s` doesn't exists" % index_name)
        ind = self.indexes_names[index_name]
        name = "%.2d%s" % (ind._order, index_name)
        name += '.py'
        with io.FileIO(os.path.join(self.path, '_indexes', name), 'r') as f:
            co = f.read()
            if code_switch == 'All':
                return str(co, 'utf-8')

            if code_switch == 'S':
                try:
                    ind = co.index(b'#SIMPLIFIED CODE')
                except ValueError:
                    return u" "
                else:
                    s = co[ind:]
                    l = s.splitlines()[1:-2]
                    ll = [x[1:] for x in l]
                    return str(b'\n'.join(ll), 'utf-8')
            if code_switch == 'P':
                try:
                    ind = co.index(b'#SIMPLIFIED CODE')
                except ValueError:
                    return str(co, 'utf-8')
                else:
                    return str(co[:ind], 'utf-8')

        return u""  # shouldn't happen

    def __set_main_storage(self):
        """
        Sets database main storage (from the **id** index)
        """
        try:
            self.storage = self.indexes_names['id'].storage
            self.id_ind = self.indexes_names['id']
        except KeyError:
            # when opening / initializing DB without `id` index
            # happens mostly on server side
            pass
            # why this line?
            # raise IndexError("Can't set main storage index. 'id' index name not found")

    def initialize(self, path=None, makedir=True):
        """
        Initialize new database

        :param path: Path to a database (allows delayed path configuration), if not provided self.path will be used
        :param makedir: Make the ``_indexes`` directory or not

        :returns: the database path
        """
        if self.opened is True:
            raise DatabaseConflict("Already opened")
        if not path:
            path = self.path
        else:
            self.path = path
        if makedir:
            if not self.path:
                raise PreconditionsException("No path specified")
            p = os.path.join(self.path, '_indexes')
            if os.path.exists(p):
                raise DatabaseConflict("Cant't create because already exists")
            os.makedirs(p)

        return self.path

    def __open_new(self, with_id_index=True, index_kwargs={}):
        """
        Will open new database (works like create),
        if not self.path provided will call initialize()
        """
        if self.path:
            if not os.path.exists(self.path):
                self.initialize(self.path)
        if not 'id' in self.indexes_names and with_id_index:
            import bitdust_forks.CodernityDB3.hash_index
            if not 'db_path' in index_kwargs:
                index_kwargs['db_path'] = self.path
            index_kwargs['name'] = 'id'
            id_ind = bitdust_forks.CodernityDB3.hash_index.UniqueHashIndex(**index_kwargs)
            self.add_index(id_ind, create=False)
            # del CodernityDB3.index
        for index in self.indexes:
            try:
                index.create_index()
            except IndexException:
                raise DatabaseConflict(
                    "Already exists (detected on index=%s)" % index.name)
        return True

    def _read_indexes(self):
        """
        Read all known indexes from ``_indexes``
        """
        p = os.path.join(self.path, '_indexes')
        for ind in os.listdir(p):
            if ind.endswith('.py'):
                self.add_index('path:' + ind, create=False)

    def __compat_things(self):
        """
        Things for compatibility.
        """
        # patch for rev size change
        if not self.id_ind:
            return
        if self.id_ind.entry_line_format[4:6] == '4s':
            # rev compatibility...
            warnings.warn("Your database is using old rev mechanizm \
for ID index. You should update that index \
(CodernityDB3.migrate.migrate).")
            from .misc import random_hex_4
            self.create_new_rev = random_hex_4

    def create(self, path=None, **kwargs):
        """
        Create database

        :param path: path where to create the database

        :returns: database path
        """
        if path:
            self.initialize(path)
        if not self.path:
            raise PreconditionsException("No path specified")
        if self.opened is True:
            raise DatabaseConflict("Already opened")
        self.__open_new(**kwargs)
        self.__set_main_storage()
        self.__compat_things()
        self.opened = True
        return self.path

    def exists(self, path=None):
        """
        Checks if database in given path exists

        :param path: path to look for database
        """
        if not path:
            path = self.path
        if not path:
            return False
        if os.path.exists(path):
            return os.path.exists(os.path.join(path, '_indexes'))
        return False

    def open(self, path=None):
        """
        Will open already existing database

        :param path: path with database to open
        """
        if self.opened is True:
            raise DatabaseConflict("Already opened")
#        else:
        if path:
            self.path = path
        if not self.path:
            raise PreconditionsException("No path specified")
        if not os.path.exists(self.path):
            raise DatabasePathException("Can't open database")
        self.indexes = []
        self.id_ind = None
        self.indexes_names = {}
        self._read_indexes()
        if not 'id' in self.indexes_names:
            raise PreconditionsException("There must be `id` index!")
        for index in self.indexes:
            index.open_index()
        self.indexes.sort(key=lambda ind: ind._order)
        self.__set_main_storage()
        self.__compat_things()
        self.opened = True
        return True

    def close(self):
        """
        Closes the database
        """
        if not self.opened:
            raise DatabaseConflict("Not opened")
        self.id_ind = None
        self.indexes_names = {}
        self.storage = None
        for index in self.indexes:
            index.close_index()
        self.indexes = []
        self.opened = False
        return True

    def destroy(self):
        """
        Allows to destroy database.

        **not reversable** operation!
        """
        # destroy all but *id*
        if not self.exists():
            raise DatabaseConflict("Doesn't exists'")
        for index in reversed(self.indexes[1:]):
            try:
                self.destroy_index(index)
            except:
                pass
        if getattr(self, 'id_ind', None) is not None:
            self.id_ind.destroy()  # now destroy id index
        # remove all files in db directory
        for root, dirs, files in os.walk(self.path, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(self.path)
        self.close()
        return True

    def _single_update_index(self, index, data, db_data, doc_id):
        """
        Performs update operation on single index

        :param index: the index to perform operation
        :param data: new data
        :param db_data: database data
        :param doc_id: the id of document
        """
        try:
            old_should_index = index.make_key_value(db_data)
        except Exception as ex:
            warnings.warn("""Problem during update for `%s`, ex = `%s`, you should check index code.""" % (index.name, ex), RuntimeWarning)
            old_should_index = None
        if old_should_index:
            old_key, old_value = old_should_index
            try:
                new_should_index = index.make_key_value(data)
            except Exception as ex:
                warnings.warn("""Problem during update for `%s`, ex = `%r`, you should check index code.""" % (index.name, ex), RuntimeWarning)
                new_should_index = None
            if new_should_index:
                new_key, new_value = new_should_index
                if new_key != old_key:
                    index.delete(doc_id, old_key)
                    index.insert_with_storage(doc_id, new_key, new_value)
                elif new_value != old_value:
                    try:
                        index.update_with_storage(doc_id, new_key, new_value)
                    except (ElemNotFound, DocIdNotFound):
                        # element should be in index but isn't
                        #(propably added new index without reindex)
                        warnings.warn("""Reindex might be required for index %s""" % index.name)
            else:
                index.delete(doc_id, old_key)
        else:  # not previously indexed
            self._single_insert_index(index, data, doc_id)

    def _update_id_index(self, _rev, data):
        """
        Performs update on **id** index
        """
        _id, value = self.id_ind.make_key_value(data)
        # print('_update_id_index', _id, value)
        db_data = self.get('id', _id)
        if db_data['_rev'] != _rev:
            raise RevConflict()
        new_rev = self.create_new_rev(_rev)
        # storage = self.storage
        # start, size = storage.update(value)
        # self.id_ind.update(_id, new_rev, start, size)
        self.id_ind.update_with_storage(_id, new_rev, value)
        return _id, new_rev, db_data

    def _update_indexes(self, _rev, data):
        """
        Performs update operation on all indexes in order
        """
        _id, new_rev, db_data = self._update_id_index(_rev, data)
        # print('_update_indexes', _id, new_rev)
        for index in self.indexes[1:]:
            self._single_update_index(index, data, db_data, _id)
        return _id, new_rev

    def _single_insert_index(self, index, data, doc_id):
        """
        Performs insert operation on single index

        :param index: index to perform operation
        :param data: new data
        :param doc_id: document id
        """
        try:
            should_index = index.make_key_value(data)
        except Exception as ex:
            warnings.warn("""Problem during insert for `%s`, ex = `%r`, you should check index code.""" % (index.name, ex), RuntimeWarning)
            should_index = None
        # print('_single_insert_index %r %r %r' % (index, should_index, doc_id))
        if should_index:
            key, value = should_index
            index.insert_with_storage(doc_id, key, value)
            # if value:
            #     storage = index.storage
            #     start, size = storage.insert(value)
            # else:
            #     start = 1
            #     size = 0
            # index.insert(doc_id, key, start, size)

    def _insert_id_index(self, _rev, data):
        """
        Performs insert on **id** index.
        """
        _id, value = self.id_ind.make_key_value(data)  # may be improved
#        storage = self.storage
        # start, size = storage.insert(value)
        # self.id_ind.insert(_id, _rev, start, size)
        self.id_ind.insert_with_storage(_id, _rev, value)
        return _id

    def _insert_indexes(self, _rev, data):
        """
        Performs insert operation on all indexes in order
        """
        _id = self._insert_id_index(_rev, data)
        # print('database._insert_indexes %r' % _id)
        for index in self.indexes[1:]:
            self._single_insert_index(index, data, _id)

    def _single_delete_index(self, index, data, doc_id, old_data):
        """
        Performs single delete operation on single index.
        It's very similar to update functions (that's why data is in arguments)

        :param index: index to perform operation
        :param data: not important (because of update operations)
        :param doc_id: document id
        :param old_data: current data in database
        """
        index_data = index.make_key_value(old_data)
        if not index_data:
            return
        key, value = index_data
        try:
            index.delete(doc_id, key)
        except TryReindexException as exc:
            # print("Error in _single_delete_index: %r" % exc)
            return

    def _delete_id_index(self, _id, _rev, data):
        """
        Performs delete from **id** index
        """
        # key, value = self.id_ind.make_key_value(data)
        # key = data['_id']
        key = self.id_ind.make_key(_id)
        self.id_ind.delete(key)

    def _delete_indexes(self, _id, _rev, data):
        """
        Performs delete operation on all indexes in order
        """
        old_data = self.get('id', _id)
        if old_data['_rev'] != _rev:
            raise RevConflict()
        for index in self.indexes[1:]:
            self._single_delete_index(index, data, _id, old_data)
        self._delete_id_index(_id, _rev, data)

    def destroy_index(self, index):
        """
        Destroys index

        :param index: the index to destroy
        :type index: :py:class:`CodernityDB3.index.Index`` instance, or string
        """
        if isinstance(index, six.string_types):
            if not index in self.indexes_names:
                raise PreconditionsException("No index named %s" % index)
            index = self.indexes_names[index]
        elif not index in self.indexes:
            self.__not_opened()
            raise PreconditionsException("Argument must be Index instance or valid string index format")
        if index.name == 'id':
            self.__not_opened()
            raise PreconditionsException("Id index cannot be destroyed")
        full_file = "%.2d%s" % (index._order, index.name) + '.py'
        p = os.path.join(self.path, '_indexes', full_file)
        os.unlink(p)
        index.destroy()
        del self.indexes_names[index.name]
        self.indexes.remove(index)

    def compact_index(self, index):
        """
        Compacts index
        Used for better utilization of index metadata.
        The deleted documents will be not more in structure.

        :param index: the index to destroy
        :type index: :py:class:`CodernityDB3.index.Index`` instance, or string
        """
        if isinstance(index, six.string_types):
            if not index in self.indexes_names:
                raise PreconditionsException("No index named %s" % index)
            index = self.indexes_names[index]
        elif not index in self.indexes:
            self.__not_opened()
            raise PreconditionsException("Argument must be Index instance or valid string index format")
        if getattr(index, 'compacting', False):
            raise ReindexException(
                "The index=%s is still compacting" % index.name)
        index.compacting = True
        index.compact()
        del index.compacting

    def _compact_indexes(self):
        """
        Runs compact on all indexes
        """
        for index in self.indexes:
            self.compact_index(index)

    def _single_reindex_index(self, index, data):
        doc_id, rev, start, size, status = self.id_ind.get(
            data['_id'])  # it's cached so it's ok
        # print(status)
        if status != b'd' and status != b'u' and status != 'd' and status != 'u':
            self._single_insert_index(index, data, doc_id)

    def reindex_index(self, index):
        """
        Performs reindex on index. Optimizes metadata and storage informations for given index.

        You can't reindex **id** index.

        :param index: the index to reindex
        :type index: :py:class:`CodernityDB3.index.Index`` instance, or string
        """
        if isinstance(index, six.string_types):
            if not index in self.indexes_names:
                raise PreconditionsException("No index named %s" % index)
            index = self.indexes_names[index]
        elif not index in self.indexes:
            self.__not_opened()
            raise PreconditionsException("Argument must be Index instance or valid string index format")
        if index.name == 'id':
            self.__not_opened()
            raise PreconditionsException("Id index cannot be reindexed")
        if getattr(index, 'reindexing', False):
            raise ReindexException(
                "The index=%s is still reindexing" % index.name)

        all_iter = self.all('id')
        index.reindexing = True
        index.destroy()
        index.create_index()

        while True:
            try:
                curr = next(all_iter)
            except StopIteration:
                break
            else:
                self._single_reindex_index(index, curr)
        del index.reindexing

    def _reindex_indexes(self):
        for index in self.indexes[1:]:
            self.reindex_index(index)

    def insert(self, data):
        """
        It's using **reference** on the given data dict object,
        to avoid it copy it before inserting!

        If data **will not** have ``_id`` field,
        it will be generated (random 32 chars string)

        :param data: data to insert
        """
        if '_rev' in data:
            self.__not_opened()
            raise PreconditionsException(
                "Can't add record with forbidden fields")
        _rev = self.create_new_rev()
        if not '_id' in data:
            try:
                _id = self.id_ind.create_key()
            except:
                self.__not_opened()
                raise DatabaseException("No id?")
        else:
            _id = data['_id']
        assert _id is not None
        data['_rev'] = _rev  # for make_key_value compat with update / delete
        data['_id'] = _id
        # print('database.insert %r %r' % (_id, _rev))
        self._insert_indexes(_rev, data)
        ret = {'_id': _id, '_rev': _rev}
        data.update(ret)
        return ret

    def update(self, data):
        """
        It's using **reference** on the given data dict object,
        to avoid it copy it before updating!

        ``data`` **must** contain ``_id`` and ``_rev`` fields.

        :param data: data to update
        """
        if not '_rev' in data or not '_id' in data:
            self.__not_opened()
            raise PreconditionsException("Can't update without _rev or _id")
        _rev = data['_rev']
        try:
            # _rev = codecs.encode(_rev)
            pass
        except:
            self.__not_opened()
            raise PreconditionsException(
                "`_rev` must be valid bytes object")
        _id, new_rev = self._update_indexes(_rev, data)
        if isinstance(_id, six.binary_type):
            _id = _id.decode()
        if isinstance(new_rev, six.binary_type):
            new_rev = new_rev.decode()
        ret = {'_id': _id, '_rev': new_rev}
        data.update(ret)
        return ret

    def get(self, index_name, key, with_doc=False, with_storage=True):
        """
        Get single data from Database by ``key``.

        :param index_name: index to get data from
        :param key: key to get
        :param with_doc: if ``True`` data from **id** index will be
            included in output
        :param with_storage: if ``True`` data from index storage will be
            included, otherwise just metadata.
        """
        # if not self.indexes_names.has_key(index_name):
        #     raise DatabaseException, "Invalid index name"
        # print('database.get %r key=%r' % (index_name, key))
        if isinstance(key, six.text_type):
            key = codecs.encode(key, 'utf-8')
        try:
            ind = self.indexes_names[index_name]
        except KeyError:
            self.__not_opened()
            raise IndexNotFoundException(
                "Index `%s` doesn't exists" % index_name)
        try:
            l_key, _unk, start, size, status = ind.get(key)
        except ElemNotFound as ex:
            raise RecordNotFound(ex)
        if not start and not size:
            raise RecordNotFound("Not found")
        elif status == b'd' or status == 'd':
            raise RecordDeleted("Deleted")
        if with_storage and size:
            storage = ind.storage
            data = storage.get(start, size, status)
        else:
            data = {}
        if with_doc and index_name != 'id':
            storage = ind.storage
            doc = self.get('id', l_key, False)
            if data:
                data['doc'] = doc
            else:
                data = {'doc': doc}
        data['_id'] = l_key
        if index_name == 'id':
            data['_rev'] = _unk
        else:
            data['key'] = _unk
        return data

    def get_many(self, index_name, key=None, limit=-1, offset=0, with_doc=False, with_storage=True, start=None, end=None, **kwargs):
        """
        Allows to get **multiple** data for given ``key`` for *Hash based indexes*.
        Also allows get **range** queries for *Tree based indexes* with ``start`` and ``end`` arguments.

        :param index_name: Index to perform the operation
        :param key: key to look for (has to be ``None`` to use range queries)
        :param limit: defines limit for query
        :param offset: defines offset (how many records from start it will ignore)
        :param with_doc: if ``True`` data from **id** index will be included in output
        :param with_storage: if ``True`` data from index storage will be included, otherwise just metadata.
        :param start: ``start`` parameter for range queries
        :param end: ``end`` parameter for range queries

        :returns: iterator over records
        """
        if index_name == 'id':
            self.__not_opened()
            raise PreconditionsException("Can't get many from `id`")
        try:
            ind = self.indexes_names[index_name]
        except KeyError:
            self.__not_opened()
            raise IndexNotFoundException(
                "Index `%s` doesn't exists" % index_name)
        storage = ind.storage
        if start is None and end is None:
            gen = ind.get_many(key, limit, offset)
        else:
            gen = ind.get_between(start, end, limit, offset, **kwargs)
        while True:
            try:
                ind_data = next(gen)
            except StopIteration:
                break
            else:
                if with_storage and ind_data[-2]:
                    data = storage.get(*ind_data[-3:])
                else:
                    data = {}
                doc_id = ind_data[0]
                if with_doc:
                    doc = self.get('id', doc_id, False)
                    if data:
                        data['doc'] = doc
                    else:
                        data = {'doc': doc}
                data['_id'] = doc_id
                if key is None:
                    data['key'] = ind_data[1]
                yield data

    def all(self, index_name, limit=-1, offset=0, with_doc=False, with_storage=True):
        """
        Alows to get all records for given index

        :param index_name: Index to perform the operation
        :param limit: defines limit for query
        :param offset: defines offset (how many records from start it will ignore)
        :param with_doc: if ``True`` data from **id** index will be included in output
        :param with_storage: if ``True`` data from index storage will be included, otherwise just metadata
        """
        try:
            ind = self.indexes_names[index_name]
        except KeyError:
            self.__not_opened()
            raise IndexNotFoundException(
                "Index `%s` doesn't exists" % index_name)
        storage = ind.storage
        gen = ind.all(limit, offset)
        while True:
            try:
                doc_id, unk, start, size, status = next(gen)
                # print('database.all %r doc_id=%r unk=%r start=%r size=%r status=%r' % (index_name, doc_id, unk, start, size, status))
            except StopIteration:
                break
            else:
                if index_name == 'id':
                    if with_storage and size:
                        data = storage.get(start, size, status)
                    else:
                        data = {}
                    data['_id'] = doc_id
                    data['_rev'] = unk
                else:
                    data = {}
                    if with_storage and size:
                        data['value'] = storage.get(start, size, status)
                    data['key'] = unk
                    data['_id'] = doc_id
                    if with_doc:
                        doc = self.get('id', doc_id, False)
                        data['doc'] = doc
                yield data

    def run(self, index_name, target_funct, *args, **kwargs):
        """
        Allows to execute given function on Database side
        (important for server mode)

        If ``target_funct==sum`` then given index must have ``run_sum`` method.

        :param index_name: index name to perform action.
        :param target_funct: target function name (without *run* prefix)
        :param \*args: ``*args`` for function
        :param \*\*kwargs: ``**kwargs`` for function

        """
        try:
            ind = self.indexes_names[index_name]
        except KeyError:
            self.__not_opened()
            raise IndexNotFoundException(
                "Index `%s` doesn't exists" % index_name)
        try:
            funct = getattr(ind, "run_" + target_funct)
        except AttributeError:
            raise IndexException("Invalid function to run")
        return funct(self, *args, **kwargs)

    def count(self, target_funct, *args, **kwargs):
        """
        Counter. Allows to execute for example

        .. code-block:: python

            db.count(db.all, 'id')

        And it will return then how much records are in your ``id`` index.

        .. warning::
            It sets ``kwargs['with_storage'] = False`` and ``kwargs['with_doc'] = False``


        """
        kwargs['with_storage'] = False
        kwargs['with_doc'] = False
        iter_ = target_funct(*args, **kwargs)
        i = 0
        while True:
            try:
                next(iter_)
                i += 1
            except StopIteration:
                break
        return i

    def delete(self, data):
        """
        Delete data from database.

        ``data`` has to contain ``_id`` and ``_rev`` fields.

        :param data: data to delete
        """
        if not '_rev' in data or not '_id' in data:
            raise PreconditionsException("Can't delete without _rev or _id")
        _id = data['_id']
        _rev = data['_rev']
        try:
            pass
        except:
            raise PreconditionsException(
                "`_id` and `_rev` must be valid bytes object")
        data['_deleted'] = True
        self._delete_indexes(_id, _rev, data)
        return True

    def compact(self):
        """
        Compact all indexes. Runs :py:meth:`._compact_indexes` behind.
        """
        self.__not_opened()
        self._compact_indexes()

    def reindex(self):
        """
        Reindex all indexes. Runs :py:meth:`._reindex_indexes` behind.
        """
        self.__not_opened()
        self._reindex_indexes()

    def flush_indexes(self):
        """
        Flushes all indexes
        """
        self.__not_opened()
        for index in self.indexes:
            index.flush()

    def flush(self):
        """
        Flushes all indexes. Runs :py:meth:`.flush_indexes` behind.
        """
        return self.flush_indexes()

    def fsync(self):
        """
        It forces the kernel buffer to be written to disk. Use when you're sure that you need to.
        """
        self.__not_opened()
        for index in self.indexes:
            index.flush()
            index.fsync()

    def __get_size(self):
        """
        :returns: total size of database.
        """
        if not self.path:
            return 0
        return sum(
            os.path.getsize(os.path.join(dirpath, filename)) for dirpath, dirnames,
            filenames in os.walk(self.path) for filename in filenames)

    def get_index_details(self, name):
        """
        Will return index properties.

        :returns: index details
        """
        self.__not_opened()
        try:
            db_index = self.indexes_names[name]
        except KeyError:
            self.__not_opened()
            raise IndexNotFoundException("Index doesn't exist")

        props = {}
        for key, value in db_index.__dict__.items():
            if not isinstance(value, collections.Callable):  # not using inspect etc...
                props[key] = value

        return props

    def get_db_details(self):
        """
        Get's database details, size, indexes, environment etc.

        :returns: database details
        """
        props = {}
        props['path'] = self.path
        props['size'] = self.__get_size()
        props['indexes'] = list(self.indexes_names.keys())
        props['cdb_environment'] = cdb_environment
        return props



#!/usr/bin/env python
# list_files_orator.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (list_files_orator.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: list_files_orator.

.. raw:: html

    <a href="https://bitdust.io/automats/list_files_orator/list_files_orator.png" target="_blank">
    <img src="https://bitdust.io/automats/list_files_orator/list_files_orator.png" style="max-width:100%;">
    </a>

This simple state machine requests a list of files stored on remote nodes.

Before that, it scans the local backup folder and prepare an index of existing data pieces.


EVENTS:
    * :red:`inbox-files`
    * :red:`init`
    * :red:`local-files-done`
    * :red:`need-files`
    * :red:`supplier-connected`
    * :red:`timer-20sec`
    * :red:`timer-2sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import time

#------------------------------------------------------------------------------

from twisted.internet.defer import maybeDeferred, Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.contacts import contactsdb

from bitdust.main import events
from bitdust.main import settings

from bitdust.services import driver

from bitdust.p2p import p2p_service
from bitdust.p2p import online_status
from bitdust.p2p import p2p_connector
from bitdust.p2p import propagate

from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_ListFilesOrator = None

#------------------------------------------------------------------------------


def is_synchronized(customer_idurl=None):
    if not A():
        return False
    customer_idurl = customer_idurl or my_id.getIDURL()
    if A().state == 'SAW_FILES':
        if A().last_time_saw_files.get(customer_idurl, -1) > 0:
            return True
        return False
    if A().state in [
        'LOCAL_FILES',
        'REMOTE_FILES',
    ]:
        if A().target_customer_idurl != customer_idurl:
            return A().last_time_saw_files.get(customer_idurl, -1) > 0
        lt_saw_files = A().last_time_saw_files.get(customer_idurl, -1)
        if lt_saw_files > 0 and time.time() - lt_saw_files < 20:
            return True
    return False


def synchronize_files(customer_idurl=None):
    ret = Deferred()
    if not A():
        ret.errback(Exception('not initialized'))
        return ret
    customer_idurl = customer_idurl or my_id.getIDURL()
    if A().state in ['SAW_FILES', 'NO_FILES']:
        A('need-files', customer_idurl=customer_idurl, result_defer=ret)
        return ret

    def _on_list_files_orator_state_changed(oldstate, newstate, event_string, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, oldstate=oldstate, newstate=newstate, event=event_string)
        if newstate != oldstate and newstate in ['SAW_FILES', 'NO_FILES']:
            A().removeStateChangedCallback(_on_list_files_orator_state_changed)
            A('need-files', customer_idurl=customer_idurl, result_defer=ret)

    A().addStateChangedCallback(_on_list_files_orator_state_changed)
    return ret


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _ListFilesOrator
    if event is None:
        return _ListFilesOrator
    if _ListFilesOrator is None:
        _ListFilesOrator = ListFilesOrator(
            name='list_files_orator',
            state='NO_FILES',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _ListFilesOrator.automat(event, *args, **kwargs)
    return _ListFilesOrator


def Destroy():
    """
    Destroy list_files_orator() automat and remove its instance from memory.
    """
    global _ListFilesOrator
    if _ListFilesOrator is None:
        return
    _ListFilesOrator.destroy()
    del _ListFilesOrator
    _ListFilesOrator = None


class ListFilesOrator(automat.Automat):

    """
    A class to request list of my files from my suppliers and also scan the
    local files.
    """

    timers = {
        'timer-2sec': (2.0, ['REMOTE_FILES']),
        'timer-20sec': (20.0, ['REMOTE_FILES']),
    }

    def init(self):
        self.target_customer_idurl = None
        self.result_defer = None
        self.critical_suppliers_number = 0
        self.last_time_saw_files = {}
        self.ping_required = True
        self.received_lf_counter = 0
        self.requested_lf_packet_ids = set()
        events.add_subscriber(self._on_my_identity_rotated, 'my-identity-rotated')
        events.add_subscriber(self._on_supplier_connected, 'supplier-connected')

    def shutdown(self):
        self.result_defer = None
        self.received_lf_counter = 0
        self.last_time_saw_files.clear()
        self.requested_lf_packet_ids.clear()
        events.remove_subscriber(self._on_supplier_connected, 'supplier-connected')
        events.remove_subscriber(self._on_my_identity_rotated, 'my-identity-rotated')

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        if self.target_customer_idurl is None or self.target_customer_idurl == my_id.getIDURL():
            if driver.is_on('service_backups'):
                # TODO: rebuild using "list-files-orator-state-changed" event
                from bitdust.storage import backup_monitor
                backup_monitor.A('list_files_orator.state', newstate)
        if newstate == 'SAW_FILES':
            lt_saw_files = self.last_time_saw_files.get(self.target_customer_idurl, -1)
            if lt_saw_files <= 0 or time.time() - lt_saw_files < 20:
                events.send('my-list-files-refreshed', data={'customer_idurl': self.target_customer_idurl})
            self.last_time_saw_files[self.target_customer_idurl] = time.time()
        if newstate == 'NO_FILES':
            self.last_time_saw_files[self.target_customer_idurl] = -1
        if newstate in ['SAW_FILES', 'NO_FILES']:
            self.target_customer_idurl = None
            if self.result_defer:
                if not self.result_defer.called:
                    self.result_defer.callback(self.state)
                self.result_defer = None

    def A(self, event, *args, **kwargs):
        #---NO_FILES---
        if self.state == 'NO_FILES':
            if event == 'need-files':
                self.state = 'LOCAL_FILES'
                self.doReadLocalFiles(*args, **kwargs)
            elif event == 'init':
                pass
        #---LOCAL_FILES---
        elif self.state == 'LOCAL_FILES':
            if event == 'local-files-done' and p2p_connector.A().state == 'CONNECTED':
                self.state = 'REMOTE_FILES'
                self.doRequestFilesAllSuppliers(*args, **kwargs)
            elif event == 'local-files-done' and p2p_connector.A().state != 'CONNECTED':
                self.state = 'NO_FILES'
        #---REMOTE_FILES---
        elif self.state == 'REMOTE_FILES':
            if event == 'supplier-connected':
                self.doRequestFilesOneSupplier(*args, **kwargs)
            elif event == 'timer-20sec' and not self.isEnoughListFilesReceived(*args, **kwargs) and not self.isSomeConnecting(*args, **kwargs):
                self.state = 'NO_FILES'
            elif (event == 'timer-2sec' and self.isEnoughListFilesReceived(*args, **kwargs)) or (event == 'inbox-files' and self.isAllListFilesReceived(*args, **kwargs)):
                self.state = 'SAW_FILES'
        #---SAW_FILES---
        elif self.state == 'SAW_FILES':
            if event == 'need-files':
                self.state = 'LOCAL_FILES'
                self.doReadLocalFiles(*args, **kwargs)
        return None

    def isAllListFilesReceived(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'list_files_orator.isAllListFilesReceived need %d more' % len(self.requested_lf_packet_ids))
        return len(self.requested_lf_packet_ids) == 0

    def isSomeConnecting(self, *args, **kwargs):
        """
        Condition method.
        """
        from bitdust.customer import supplier_connector
        for one_supplier_connector in supplier_connector.connectors(customer_idurl=self.target_customer_idurl).values():
            if one_supplier_connector.state not in ['CONNECTED', 'DISCONNECTED', 'NO_SERVICE']:
                return True
        return False

    def isEnoughListFilesReceived(self, *args, **kwargs):
        """
        Condition method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'list_files_orator.isEnoughListFilesReceived %d list files received' % self.received_lf_counter)
        return self.received_lf_counter >= self.critical_suppliers_number

    def doReadLocalFiles(self, *args, **kwargs):
        from bitdust.storage import backup_matrix
        self.target_customer_idurl = kwargs.get('customer_idurl') or my_id.getIDURL()
        self.result_defer = kwargs.get('result_defer')
        self.critical_suppliers_number = 0
        maybeDeferred(backup_matrix.ReadLocalFiles).addBoth(lambda x: self.automat('local-files-done'))

    def doRequestFilesAllSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        if self.ping_required:
            self.ping_required = False
            propagate.ping_suppliers(customer_idurl=self.target_customer_idurl).addBoth(self._do_request)
        else:
            self._do_request()

    def doRequestFilesOneSupplier(self, *args, **kwargs):
        """
        Action method.
        """
        supplier_idurl = args[0]
        if _Debug:
            lg.out(_DebugLevel, 'list_files_orator.doRequestFilesOneSupplier from %s' % supplier_idurl)
        outpacket = p2p_service.SendListFiles(
            target_supplier=supplier_idurl,
            customer_idurl=self.target_customer_idurl,
            timeout=settings.P2PTimeOut(),
        )
        if outpacket:
            self.requested_lf_packet_ids.add(outpacket.PacketID)
        else:
            lg.err('failed sending ListFiles() to %r' % supplier_idurl)

    def _do_request(self, x=None):
        from bitdust.raid import eccmap
        self.received_lf_counter = 0
        self.requested_lf_packet_ids.clear()
        known_suppliers = contactsdb.suppliers(customer_idurl=self.target_customer_idurl)
        try:
            self.critical_suppliers_number = eccmap.GetCorrectableErrors(len(known_suppliers))
        except:
            lg.warn('number of known suppliers for customer %r is not standard' % self.target_customer_idurl)
            self.critical_suppliers_number = int(float(len(known_suppliers))*0.75)
        for idurl in known_suppliers:
            if idurl:
                if online_status.isOnline(idurl):
                    if _Debug:
                        lg.out(_DebugLevel, 'list_files_orator._do_request  ListFiles() from my supplier %s' % idurl)
                    outpacket = p2p_service.SendListFiles(
                        target_supplier=idurl,
                        customer_idurl=self.target_customer_idurl,
                        timeout=settings.P2PTimeOut(),
                    )
                    if outpacket:
                        self.requested_lf_packet_ids.add(outpacket.PacketID)
                    else:
                        lg.err('failed sending ListFiles() to %r' % idurl)
                else:
                    lg.warn('skip sending ListFiles() because %s is not online' % idurl)

    def _on_my_identity_rotated(self, evt):
        self.ping_required = True
        if _Debug:
            lg.dbg(_DebugLevel, 'updating ping_required=True')

    def _on_supplier_connected(self, evt):
        if id_url.field(evt.data['customer_idurl']) == self.target_customer_idurl:
            if _Debug:
                lg.dbg(_DebugLevel, 'for customer %r single supplier %r was connected' % (self.target_customer_idurl, evt.data['supplier_idurl']))
            self.automat('supplier-connected', evt.data['supplier_idurl'])


#------------------------------------------------------------------------------


def IncomingListFiles(newpacket):
    """
    Called from ``p2p.backup_control`` to pass incoming "ListFiles" packet
    here.
    """
    if not A():
        return
    if newpacket.PacketID in A().requested_lf_packet_ids:
        A().received_lf_counter += 1
        A().requested_lf_packet_ids.discard(newpacket.PacketID)
        A('inbox-files', newpacket)
    else:
        if _Debug:
            lg.dbg(_DebugLevel, 'received and ignored %r, currently target customer is %r' % (newpacket, A().target_customer_idurl))



#!/usr/bin/python
# service_proxy_transport.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_proxy_transport.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_proxy_transport
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return ProxyTransportService()


class ProxyTransportService(LocalService):

    service_name = 'service_proxy_transport'
    config_path = 'services/proxy-transport/enabled'
    proto = 'proxy'
    transport = None
    stop_when_failed = True

    def init(self):
        self.starting_deferred = None

    def dependent_on(self):
        depends = [
            'service_identity_propagate',
            'service_nodes_lookup',
        ]
        depends.extend(self._available_transports())
        return depends

    def start(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.transport.proxy import proxy_interface
        from bitdust.transport import network_transport
        from bitdust.transport import gateway
        from bitdust.services import driver
        from bitdust.main import events
        from bitdust.main.config import conf
        if len(self._available_transports()) == 0:
            lg.warn('no transports available')
            return False
        events.add_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        self._check_reset_original_identity()
        self.starting_deferred = Deferred()
        self.transport = network_transport.NetworkTransport('proxy', proxy_interface.GateInterface())
        conf().addConfigNotifier('services/proxy-transport/enabled', self._on_enabled_disabled)
        conf().addConfigNotifier('services/proxy-transport/sending-enabled', self._on_sending_enabled_disabled)
        conf().addConfigNotifier('services/proxy-transport/receiving-enabled', self._on_receiving_enabled_disabled)
        if driver.is_on('service_entangled_dht'):
            self._do_join_proxy_routers_dht_layer()
        else:
            self.transport.automat('init', (gateway.listener(), self._on_transport_state_changed))
            reactor.callLater(0, self.transport.automat, 'start')  # @UndefinedVariable
        return self.starting_deferred

    def stop(self):
        from twisted.internet.defer import succeed
        from bitdust.main import events
        from bitdust.main.config import conf
        events.remove_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        conf().removeConfigNotifier('services/proxy-transport/enabled')
        conf().removeConfigNotifier('services/proxy-transport/sending-enabled')
        conf().removeConfigNotifier('services/proxy-transport/receiving-enabled')
        t = self.transport
        self.transport = None
        t.automat('shutdown')
        return succeed(True)

    def installed(self):
        try:
            from bitdust.transport.proxy import proxy_interface
        except:
            from bitdust.logs import lg
            lg.exc()
            return False
        return True

    def health_check(self):
        from bitdust.services import driver
        all_deps_are_ok = True
        for dep_name in self.dependent_on():
            if not driver.is_enabled(dep_name):
                all_deps_are_ok = False
                break
            if not driver.is_on(dep_name):
                all_deps_are_ok = False
                break
        if not all_deps_are_ok:
            return False
        return self.state == 'ON'

    def _available_transports(self):
        from bitdust.main import settings
        atransports = []
        if settings.enableTCP() and settings.enableTCPreceiving() and settings.enableTCPsending():
            atransports.append('service_tcp_transport')
        if settings.enableUDP() and settings.enableUDPreceiving() and settings.enableUDPsending():
            atransports.append('service_udp_transport')
        return atransports

    def _reset_my_original_identity(self):
        from bitdust.userid import my_id
        from bitdust.main.config import conf
        from bitdust.logs import lg
        lg.warn('RESET my-original-identity')
        conf().setData('services/proxy-transport/my-original-identity', '')
        conf().setString('services/proxy-transport/current-router', '')
        my_id.rebuildLocalIdentity()

    def _check_reset_original_identity(self):
        from bitdust.logs import lg
        from bitdust.lib import misc
        from bitdust.lib import strng
        from bitdust.main.config import conf
        from bitdust.userid import identity
        from bitdust.userid import my_id
        from bitdust.userid import id_url
        orig_ident_xmlsrc = conf().getData('services/proxy-transport/my-original-identity', '').strip()
        current_router_idurl = conf().getString('services/proxy-transport/current-router', '').strip()
        if current_router_idurl:
            current_router_idurl = id_url.field(current_router_idurl.split(' ')[0])
        if not orig_ident_xmlsrc:
            if current_router_idurl:
                lg.warn('"current-router" is %s, but "my-original-identity" is empty' % current_router_idurl)
            else:
                lg.warn('"current-router" and "my-original-identity" is empty')
            self._reset_my_original_identity()
            return
        orig_ident = identity.identity(xmlsrc=orig_ident_xmlsrc)
        if not orig_ident.isCorrect() or not orig_ident.Valid():
            lg.warn('"my-original-identity" config has not valid value')
            self._reset_my_original_identity()
            return
        if orig_ident.getIDURL() != my_id.getIDURL():
            lg.warn('"my-original-identity" source is not equal to local identity source')
            self._reset_my_original_identity()
            return
        externalIP = strng.to_bin(misc.readExternalIP())
        if externalIP and strng.to_bin(orig_ident.getIP()) != externalIP:
            lg.warn('external IP was changed : reset "my-original-identity" config')
            self._reset_my_original_identity()
            return
        if not current_router_idurl:
            lg.warn('"my-original-identity" config is correct, but current router is empty')
            self._reset_my_original_identity()
        all_orig_contacts_present_in_local_identity = True
        for orig_contact in orig_ident.getContacts():
            if orig_contact not in my_id.getLocalIdentity().getContacts():
                all_orig_contacts_present_in_local_identity = False
        if all_orig_contacts_present_in_local_identity:
            lg.warn('all of "my-original-identity" contacts is found in local identity: need to RESET!')
            self._reset_my_original_identity()

    def _do_join_proxy_routers_dht_layer(self):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.dht import known_nodes
        lg.info('going to join proxy routers DHT layer: %d' % dht_records.LAYER_PROXY_ROUTERS)
        known_seeds = known_nodes.nodes()
        d = dht_service.open_layer(
            layer_id=dht_records.LAYER_PROXY_ROUTERS,
            seed_nodes=known_seeds,
            connect_now=True,
            attach=False,
        )
        d.addCallback(self._on_proxy_routers_dht_layer_connected)
        d.addErrback(self._on_proxy_routers_dht_layer_connect_failed)

    def _on_proxy_routers_dht_layer_connected(self, ok):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.transport import gateway
        from bitdust.userid import my_id
        lg.info('connected to DHT layer for proxy routers: %r' % ok)
        if my_id.getIDURL():
            dht_service.set_node_data('idurl', my_id.getIDURL().to_text(), layer_id=dht_records.LAYER_PROXY_ROUTERS)
        if self.transport:
            if self.starting_deferred and not self.starting_deferred.called:
                self.transport.automat('init', (gateway.listener(), self._on_transport_state_changed))
                reactor.callLater(0, self.transport.automat, 'start')  # @UndefinedVariable
        return ok

    def _on_proxy_routers_dht_layer_connect_failed(self, err):
        from bitdust.logs import lg
        from bitdust.transport import gateway
        lg.err('failed to connect to DHT layer: %r' % err)
        if self.starting_deferred and not self.starting_deferred.called:
            self.transport.automat('init', (gateway.listener(), self._on_transport_state_changed))
            reactor.callLater(0, self.transport.automat, 'start')  # @UndefinedVariable

    def _on_transport_state_changed(self, transport, oldstate, newstate):
        from bitdust.logs import lg
        from bitdust.p2p import p2p_connector
        lg.info('%s -> %s in %r  starting_deferred=%r' % (oldstate, newstate, transport, bool(self.starting_deferred)))
        if self.starting_deferred:
            if newstate == 'LISTENING' and oldstate != newstate:
                self.starting_deferred.callback(True)
                self.starting_deferred = None
                p2p_connector.A('check-synchronize')
            if newstate == 'OFFLINE' and oldstate != newstate and oldstate not in [
                'INIT',
            ]:
                self.starting_deferred.errback(Exception(newstate))
                self.starting_deferred = None
                p2p_connector.A('check-synchronize')

    def _on_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        network_connector.A('reconnect')

    def _on_receiving_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        network_connector.A('reconnect')

    def _on_sending_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        network_connector.A('reconnect')

    def _on_dht_layer_connected(self, evt):
        from bitdust.dht import dht_records
        if evt.data['layer_id'] == 0:
            self._do_join_proxy_routers_dht_layer()
        elif evt.data['layer_id'] == dht_records.LAYER_PROXY_ROUTERS:
            self._on_proxy_routers_dht_layer_connected(True)



WEB_ROOT_TEMPLATE = """<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="msapplication-TileColor" content="#da532c">
    <meta name="theme-color" content="#ffffff">
    <meta property="og:image:width" content="2400">
    <meta property="og:image:height" content="1257">
    <meta property="og:title" content="%(title)s">
    <meta property="og:description"
          content="Distributed secure anonymous on-line storage, where only the owner has access and absolute control over its data.">
    <meta property="og:url" content="%(site_url)s">
    <meta property="og:image" content="%(basepath)sassets/img/og/og-image.jpg">

    <title data-content="pageTitle">%(title)s</title>

    <link rel="manifest" href="%(basepath)ssite.webmanifest"/>

    <link rel="apple-touch-icon" sizes="180x180" href="%(basepath)slogos/logo-pictogram-color.png"/>
    <link rel="icon" type="image/png" sizes="32x32" href="%(basepath)slogos/logo-pictogram-color.png"/>
    <link rel="icon" type="image/png" sizes="16x16" href="%(basepath)slogos/logo-pictogram-color.png"/>
    <link rel="mask-icon" href="%(basepath)slogos/logo-pictogram-color.svg" color="#5bbad5"/>

    <link href="https://fonts.googleapis.com/css?family=Source+Sans+Pro:600,700" rel="stylesheet">
    <link href="https://fonts.googleapis.com/css?family=Open+Sans:600" rel="stylesheet">
    <link rel="stylesheet" href="%(basepath)sassets/css/applify.min.css" type="text/css"/>
    <link rel="stylesheet" href="%(basepath)sassets/css/tables.css" type="text/css"/>
    <link rel="stylesheet" href="%(basepath)scss/automat.css" type="text/css"/>

</head>

<body data-fade_in="on-load">

<nav class="navbar navbar-fixed-top navbar-dark bg-indigo">
    <div class="container">

        <!-- Navbar Logo -->
        <a class="ui-variable-logo navbar-brand" href="%(site_url)s/" title="BitDust">
            <!-- Default Logo -->
            <img class="logo-default" src="%(basepath)sassets/img/logo/bitdust-logo-white.svg" alt="BitDust">
            <!-- Transparent Logo -->
            <img class="logo-transparent" src="%(basepath)sassets/img/logo/bitdust-logo-white.svg" alt="BitDust">
        </a><!-- .navbar-brand -->

        <!-- Navbar Navigation -->
        <div class="ui-navigation navbar-center">
            <ul class="nav navbar-nav">
                <!-- Nav Item -->
                <li>
                    <a href="%(site_url)s/index.html#product">Product</a>
                </li>
                <!-- Nav Item -->
                <li>
                    <a href="%(site_url)s/index.html#problem">Problem</a>
                </li>
                <!-- Nav Item -->
                <li>
                    <a href="%(site_url)s/index.html#solution">Solution</a>
                </li>
                <!-- Nav Item -->
                <li>
                    <a href="%(site_url)s/index.html#ecosystem">Ecosystem</a>
                </li>
                <!-- Nav Item -->
                <!--
                <li>
                    <a href="%(site_url)s/index.html#roadmap">Roadmap</a>
                </li>
                -->
                <!-- Nav Item -->
                <!--
                <li>
                    <a href="%(site_url)s/index.html#team">Team</a>
                </li>
                -->
                <!-- Nav Item -->
                <li class="active">
                    <a href="%(wikipath)s">Wiki</a>
                </li>

                <!-- Nav Item -->
                <li class="dropdown active">
                    <a href="#" class="dropdown-toggle" data-toggle="dropdown">
                        Network
                    </a>
                    <ul class="dropdown-menu">
                        <li class="dropdown-item">
                            <a href="%(idserverspath)s" style="color:#8089ff;">ID servers</a>
                        </li>
                        <li class="dropdown-item">
                            <a href="%(blockchainpath)s" style="color:#8089ff;">Blockchain</a>
                        </li>
                    </ul>
                </li>

            </ul><!--.navbar-nav -->
        </div><!--.ui-navigation -->

        <!-- Navbar Button -->
        <a href="%(site_url)s/index.html#product"
           class="btn btn-sm ui-gradient-peach pull-right">
            Get Started
        </a>

        <!-- Navbar Button -->
        <a href="https://forms.gle/RkJW3L7NTamZ5K8m6" target="_blank"
           class="btn btn-sm ui-gradient-peach pull-right">
            Give Feedback
        </a>

    </div><!-- .container -->
</nav> <!-- nav -->


<!-- Main -->
<div class="%(div_main_class)s" role="main">

%(div_main_body)s

<!-- Footer -->
<footer class="ui-footer bg-gray">

    %(pre_footer)s

    <!-- Footer Copyright -->
    <div class="footer-copyright bg-dark-gray">
        <div class="container">
            <div class="row">
                <!-- Copyright -->
                <div class="col-sm-6 center-on-sm">
                    <p>
                        &copy; 2019 <a href="%(site_url)s" target="_blank" title="BitDust">BitDust</a>
                    </p>
                </div>
                <!-- Social Icons -->
                <div class="col-sm-6 text-right">
                    <ul class="footer-nav">
                        <li>
                            <a href="%(wikipath)s">
                                Wiki
                            </a>
                        </li>
                        <li>
                            <a href="https://github.com/bitdust-io/">
                                Github
                            </a>
                        </li>
                    </ul>
                </div>
            </div>
        </div><!-- .container -->
    </div><!-- .footer-copyright -->

</footer><!-- .ui-footer -->

</div>
<!-- Main -->


<!-- Scripts -->
<script src="%(basepath)sassets/js/libs/jquery/jquery-3.2.1.min.js"></script>
<script src="%(basepath)sassets/js/libs/bootstrap.js"></script>
<script src="%(basepath)sassets/js/libs/owl.carousel/owl.carousel.min.js"></script>
<script src="%(basepath)sassets/js/libs/slider-pro/jquery.sliderPro.min.js"></script>
<script src="%(basepath)sassets/js/libs/form-validator/form-validator.min.js"></script>
<script src="%(basepath)sassets/js/applify/ui-map.js"></script>
<script src="%(basepath)sassets/js/applify/build/applify.js"></script>
%(google_analytics)s
<!-- Scripts -->

</body>
</html>"""

div_main_class = 'main wiki'

div_main_body = ''

GOOGLE_ANALITICS = """
<!-- Global site tag (gtag.js) - Google Analytics -->
<script async src="https://www.googletagmanager.com/gtag/js?id=%(google_analytics_code)s"></script>
<script>
    window.dataLayer = window.dataLayer || [];

    function gtag() {
        dataLayer.push(arguments);
    }

    gtag('js', new Date());

    gtag('config', '%(google_analytics_code)s');
</script>
"""

google_analytics = GOOGLE_ANALITICS % dict(google_analytics_code='', )

pre_footer = """
<div class="container pt-6 pb-6">
    <div class="row">
        <div class="col-sm-8 footer-about footer-col center-on-sm">
            <img src="%(basepath)sassets/img/logo/bitdust-logo-white.svg" alt="BitDust" />
            <p class="mt-1">
                We are currently with a small team primarily focusing on the development and supporting the BitDust community.
                Join us via Telegram or send a email.
            </p>
        </div>

        <div class="col-md-4 col-sm-6 footer-col center-on-sm">
            <div>
                <a class="btn ui-gradient-purple btn-circle shadow-md" href="https://t.me/bitdust" target="_blank">
                    <span class="fa fa-telegram"></span>
                </a>
                <a class="btn ui-gradient-blue btn-circle shadow-md" href="https://github.com/bitdust-io/" target="_blank">
                    <span class="fa fa-github"></span>
                </a>
            </div>
        </div>
    </div><!-- .row -->
</div><!-- .container -->
"""



#!/usr/bin/env python
# tcp_stream.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tcp_stream.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
..module:: tcp_stream
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open
from io import BytesIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import time
import struct
import random

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet import defer, interfaces

from zope.interface import implementer

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import tmpfile

from bitdust.main import settings

from bitdust.lib import misc
from bitdust.lib import strng

#------------------------------------------------------------------------------

MIN_PROCESS_STREAMS_DELAY = 0.1
MAX_PROCESS_STREAMS_DELAY = 5

MAX_SIMULTANEOUS_OUTGOING_FILES = 20

#------------------------------------------------------------------------------

_LastFileID = None
_ProcessStreamsDelay = 0.1
_ProcessStreamsTask = None
_StreamCounter = 0

#------------------------------------------------------------------------------


def start_process_streams():
    reactor.callLater(0, process_streams)  # @UndefinedVariable
    return True


def stop_process_streams():
    global _ProcessStreamsTask
    if _ProcessStreamsTask:
        if _ProcessStreamsTask.active():
            _ProcessStreamsTask.cancel()
        _ProcessStreamsTask = None
        return True
    return False


def process_streams():
    from bitdust.transport.tcp import tcp_node
    global _ProcessStreamsDelay
    global _ProcessStreamsTask
    has_activity = False
    for connections in tcp_node.opened_connections().values():
        for connection in connections:
            has_timeouts = False  # connection.stream.timeout_incoming_files()
            has_sends = False  # connection.stream.process_sending_data()
            has_outbox = connection.process_outbox_queue()
            if has_timeouts or has_sends or has_outbox:
                has_activity = True
    _ProcessStreamsDelay = misc.LoopAttenuation(
        _ProcessStreamsDelay,
        has_activity,
        MIN_PROCESS_STREAMS_DELAY,
        MAX_PROCESS_STREAMS_DELAY,
    )
    # attenuation
    _ProcessStreamsTask = reactor.callLater(  # @UndefinedVariable
        _ProcessStreamsDelay,
        process_streams,
    )


#------------------------------------------------------------------------------


def list_input_streams(sorted_by_time=True):
    from bitdust.transport.tcp import tcp_node
    streams = []
    for connections in tcp_node.opened_connections().values():
        for connection in connections:
            if connection.stream:
                streams.extend(list(connection.stream.inboxFiles.values()))
    if sorted_by_time:
        streams.sort(key=lambda stream: stream.started)
    return streams


def list_output_streams(sorted_by_time=True):
    from bitdust.transport.tcp import tcp_node
    streams = []
    for connections in tcp_node.opened_connections().values():
        for connection in connections:
            if connection.stream:
                streams.extend(list(connection.stream.outboxFiles.values()))
    if sorted_by_time:
        streams.sort(key=lambda stream: stream.started)
    return streams


def find_stream(file_id=None, transfer_id=None):
    from bitdust.transport.tcp import tcp_node
    for connections in tcp_node.opened_connections().values():
        for connection in connections:
            if connection.stream:
                for out_file in connection.stream.outboxFiles.values():
                    if file_id and out_file.file_id == file_id:
                        return out_file
                    if transfer_id and out_file.transfer_id == transfer_id:
                        return out_file
                for in_file in connection.stream.inboxFiles.values():
                    if file_id and in_file.file_id == file_id:
                        return in_file
                    if transfer_id and in_file.transfer_id == transfer_id:
                        return in_file
    return None


#------------------------------------------------------------------------------


def make_stream_id():
    global _StreamCounter
    _StreamCounter += 1
    return random.randint(10, 99)*1000000 + random.randint(10, 99)*10000 + _StreamCounter % 10000


def make_file_id():
    """
    Generate a unique file ID for OutboxFile.
    """
    global _LastFileID
    newid = int(str(int(time.time()*100.0))[4:])
    if _LastFileID is None:
        _LastFileID = newid
    elif _LastFileID >= newid:
        _LastFileID += 1
    else:
        _LastFileID = newid
    return _LastFileID


#------------------------------------------------------------------------------


class TCPFileStream():
    def __init__(self, connection):
        self.stream_id = make_stream_id()  # not used at the moment, use file_id instead
        self.connection = connection
        self.outboxFiles = {}
        self.inboxFiles = {}
        self.started = time.time()
        self.sender = MultipleFilesSender(self.connection.transport)

    def close(self):
        """
        """
        self.sender.close()
        self.sender = None
        self.connection = None

    def abort_files(self, reason='connection closed'):
        from bitdust.transport.tcp import tcp_connection
        inbox_file_ids_to_remove = [fid for fid in self.inboxFiles.keys()]
        outbox_file_ids_to_remove = [fid for fid in self.outboxFiles.keys()]
        if _Debug:
            lg.args(_DebugLevel, reason=reason, inbox_file_ids_to_remove=inbox_file_ids_to_remove, outbox_file_ids_to_remove=outbox_file_ids_to_remove)
        for file_id in inbox_file_ids_to_remove:
            # self.send_data(tcp_connection.CMD_ABORT, struct.pack('i', file_id) + b' ' + strng.to_bin(reason))
            self.inbox_file_done(file_id, 'failed', reason)
        for file_id in outbox_file_ids_to_remove:
            if self.sender.is_sending(file_id):
                self.sender.stopFileTransfer(file_id, reason=reason)
                if reason == 'disconnecting':
                    self.send_data(tcp_connection.CMD_ABORT, struct.pack('i', file_id) + b' ' + strng.to_bin(reason))
            self.outbox_file_done(file_id, 'failed', reason)

    def data_received(self, payload):
        """
        """
        from bitdust.transport.tcp import tcp_connection
        inp = BytesIO(payload)
        try:
            file_id = int(struct.unpack('i', inp.read(4))[0])
            file_size = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            lg.exc()
            return
        inp_data = inp.read()
        inp.close()
        if file_id not in self.inboxFiles:
            if len(self.inboxFiles) >= 2*MAX_SIMULTANEOUS_OUTGOING_FILES:
                # too many incoming files, seems remote guy is cheating - drop
                # that session!
                lg.warn('too many incoming files, close connection %s' % str(self.connection))
                self.connection.automat('disconnect')
                return
            self.create_inbox_file(file_id, file_size)
        self.inboxFiles[file_id].input_data(inp_data)
        if self.inboxFiles[file_id].is_done():
            self.send_data(tcp_connection.CMD_OK, struct.pack('i', file_id))
            self.inbox_file_done(file_id, 'finished')

    def ok_received(self, payload):
        inp = BytesIO(payload)
        try:
            file_id = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            lg.exc()
            return
        inp.close()
        if file_id not in self.outboxFiles:
            lg.warn('did not found %r in outboxFiles: %r' % (file_id, self.outboxFiles.keys()))
            return
        self.outboxFiles[file_id].ok_received = True
        if _Debug:
            lg.args(_DebugLevel*2, file_id)
        if not self.outboxFiles[file_id].registration:
            self.outbox_file_done(file_id, 'finished')

    def abort_received(self, payload):
        inp = BytesIO(payload)
        try:
            file_id = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            lg.exc()
            return
        reason = inp.read()
        inp.close()
        self.inbox_file_done(file_id, 'failed', reason)

    def send_data(self, command, payload):
        return self.connection.sendData(command, payload)

    def create_inbox_file(self, file_id, file_size):
        from bitdust.transport.tcp import tcp_interface
        infile = InboxFile(self, file_id, file_size)
        d = tcp_interface.interface_register_file_receiving(self.connection.getAddress(), self.connection.peer_idurl, infile.filename)
        d.addCallback(self.on_inbox_file_registered, file_id)
        d.addErrback(self.on_inbox_file_register_failed, file_id)
        infile.registration = d
        self.inboxFiles[file_id] = infile

    def on_inbox_file_registered(self, response, file_id):
        try:
            transfer_id = int(response)
        except:
            transfer_id = None
        self.inboxFiles[file_id].transfer_id = transfer_id
        self.inboxFiles[file_id].registration = None
        if self.inboxFiles[file_id].is_done():
            infile = self.inboxFiles[file_id]
            self.close_inbox_file(file_id)
            self.report_inbox_file(infile.transfer_id, 'finished', infile.get_bytes_received())

    def on_inbox_file_register_failed(self, err, file_id):
        lg.warn('failed to register file_id=%r session=%r err: %s' % (file_id, self.session, str(err)))
        self.connection.automat('disconnect')

    def create_outbox_file(
        self,
        filename,
        filesize,
        description,
        result_defer,
        keep_alive,
    ):
        from bitdust.transport.tcp import tcp_interface
        file_id = make_file_id()
        outfile = OutboxFile(self, filename, file_id, filesize, description, result_defer, keep_alive)
        if keep_alive:
            d = tcp_interface.interface_register_file_sending(self.connection.getAddress(), self.connection.peer_idurl, filename, description)
            d.addCallback(self.on_outbox_file_registered, file_id)
            d.addErrback(self.on_outbox_file_register_failed, file_id)
            outfile.registration = d
        self.outboxFiles[file_id] = outfile
        outfile.start()

    def on_outbox_file_registered(self, response, file_id):
        try:
            transfer_id = int(response)
        except:
            transfer_id = None
        self.outboxFiles[file_id].transfer_id = transfer_id
        self.outboxFiles[file_id].registration = None
        if self.outboxFiles[file_id].is_done():
            self.outbox_file_done(file_id, 'finished')

    def on_outbox_file_register_failed(self, err, file_id):
        lg.warn('failed to register file_id=%r connection=%r err: %s' % (str(file_id), self.connection, str(err)))
        self.connection.automat('disconnect')

    def close_outbox_file(self, file_id):
        if file_id in self.outboxFiles:
            self.outboxFiles[file_id].close()
            self.outboxFiles.pop(file_id, None)
        else:
            lg.warn('outgoing TCP file %s not exist' % file_id)

    def close_inbox_file(self, file_id):
        if self.inboxFiles.get(file_id):
            self.inboxFiles[file_id].close()
            self.inboxFiles.pop(file_id, None)
        else:
            lg.warn('incoming TCP file %s not exist' % file_id)

    def report_outbox_file(self, transfer_id, status, bytes_sent, error_message=None):
        from bitdust.transport.tcp import tcp_interface
        tcp_interface.interface_unregister_file_sending(transfer_id, status, bytes_sent, error_message)

    def report_inbox_file(self, transfer_id, status, bytes_received, error_message=None):
        from bitdust.transport.tcp import tcp_interface
        tcp_interface.interface_unregister_file_receiving(transfer_id, status, bytes_received, error_message)

    def inbox_file_done(self, file_id, status, error_message=None):
        if _Debug:
            lg.args(_DebugLevel*2, file_id, status, error_message)
        if file_id not in self.inboxFiles:
            lg.warn('file_id=%r not exist' % file_id)
            return
        infile = self.inboxFiles[file_id]
        if infile.registration:
            return
        self.close_inbox_file(file_id)
        if infile.transfer_id:
            self.report_inbox_file(infile.transfer_id, status, infile.get_bytes_received(), error_message)
        else:
            lg.warn('transfer_id is None, file_id=%r' % file_id)
        del infile

    def outbox_file_done(self, file_id, status, error_message=None):
        if _Debug:
            lg.args(_DebugLevel*2, file_id, status, error_message)
        if file_id not in self.outboxFiles:
            lg.warn('file_id=%r not exist' % file_id)
            return
        outfile = self.outboxFiles[file_id]
        if outfile.result_defer:
            outfile.result_defer.callback((outfile, status, error_message))
            outfile.result_defer = None
        if outfile.registration:
            return
        self.close_outbox_file(file_id)
        if outfile.transfer_id:
            self.report_outbox_file(outfile.transfer_id, status, outfile.get_bytes_sent(), error_message)
        if not outfile.keep_alive and not self.connection.factory.keep_alive:
            self.connection.automat('disconnect')
        del outfile


#------------------------------------------------------------------------------


class InboxFile():
    def __init__(self, stream, file_id, file_size):
        self.typ = 'tcp-in'
        self.transfer_id = None
        self.registration = None
        self.stream = stream
        self.file_id = file_id
        self.size = file_size
        self.fin, self.filename = tmpfile.make('tcp-in', extension='.tcp')
        self.bytes_received = 0
        self.started = time.time()
        self.last_block_time = time.time()
        self.timeout = max(int(self.size/settings.SendingSpeedLimit()), 3)
        if _Debug:
            lg.out(_DebugLevel, '<<<TCP-IN %s with %d bytes write to %s' % (self.file_id, self.size, self.filename))

    def close(self):
        if _Debug:
            lg.out(_DebugLevel, '<<<TCP-IN %s CLOSED with %s | %s' % (self.file_id, self.stream.connection.peer_address, self.stream.connection.peer_external_address))
        if self.fin:
            os.close(self.fin)
            self.fin = None
        self.stream = None

    def get_bytes_received(self):
        return self.bytes_received

    def input_data(self, data):
        os.write(self.fin, data)
        self.bytes_received += len(data)
        self.stream.connection.total_bytes_received += len(data)
        self.last_block_time = time.time()

    def is_done(self):
        return self.bytes_received == self.size

    def is_timed_out(self):
        return time.time() - self.started > self.timeout


#------------------------------------------------------------------------------


class OutboxFile():
    def __init__(self, stream, filename, file_id, filesize, description='', result_defer=None, keep_alive=True):
        self.typ = 'tcp-out'
        self.transfer_id = None
        self.registration = None
        self.stream = stream
        self.file_id = file_id
        self.filename = filename
        self.size = filesize
        self.description = description
        self.keep_alive = keep_alive
        self.result_defer = result_defer
        self.ok_received = False
        self.bytes_sent = 0
        self.bytes_out = 0
        self.started = time.time()
        self.timeout = max(int(self.size/settings.SendingSpeedLimit()), 6)
        self.fout = open(self.filename, 'rb')
        if _Debug:
            lg.out(_DebugLevel, '>>>TCP-OUT %s with %d bytes reading from %s' % (self.file_id, self.size, self.filename))

    def close(self):
        if _Debug:
            lg.out(_DebugLevel, '>>>TCP-OUT %s CLOSED with %s | %s' % (self.file_id, self.stream.connection.peer_address, self.stream.connection.peer_external_address))
        self.stop()
        if self.fout:
            self.fout.close()
            self.fout = None
        self.stream = None
        self.result_defer = None

    def start(self):
        d = self.stream.sender.startFileTransfer(self.file_id, self.fout, self.send_chunk, self.transform_data)
        d.addCallback(self.transfer_finished)
        d.addErrback(self.transfer_failed)

    def stop(self):
        if self.stream.sender.is_sending(self.file_id):
            self.stream.sender.stopFileTransfer(self.file_id)

    def cancel(self):
        if _Debug:
            lg.out(_DebugLevel, 'tcp_stream.OutboxFile.cancel timeout=%d' % self.timeout)
        self.stop()

    def get_bytes_sent(self):
        return self.bytes_sent

    def is_done(self):
        return self.get_bytes_sent() == self.size and self.ok_received

    def is_timed_out(self):
        return time.time() - self.started > self.timeout

    def transfer_finished(self, last_byte):
        if self.ok_received:
            self.stream.outbox_file_done(self.file_id, 'finished')
        else:
            if _Debug:
                lg.args(_DebugLevel*2, self.file_id)

    def transfer_failed(self, err):
        try:
            e = err.getErrorMessage()
        except:
            e = str(err)
        lg.warn('file_id=%r err: %r' % (self.file_id, e))
        self.stream.outbox_file_done(self.file_id, 'failed', e)
        return None

    def send_chunk(self, chunk):
        from bitdust.transport.tcp import tcp_connection
        self.stream.connection.sendData(tcp_connection.CMD_DATA, chunk)

    def transform_data(self, data):
        data = strng.to_bin(data)
        datalength = len(data)
        datagram = b''
        datagram += struct.pack('i', self.file_id)
        datagram += struct.pack('i', self.size)
        datagram += data
        self.bytes_sent += datalength
        self.stream.connection.total_bytes_sent += datalength
        return datagram


#------------------------------------------------------------------------------


@implementer(interfaces.IProducer)
class MultipleFilesSender:

    CHUNK_SIZE = 2**14

    def __init__(self, consumer):
        self.active_files = {}
        self.consumer = consumer
        self.consumer.registerProducer(self, False)

    def close(self):
        self.consumer.unregisterProducer()
        self.consumer = None
        self.active_files.clear()

    def is_sending(self, file_id):
        return file_id in self.active_files

    def startFileTransfer(self, file_id, file_object, writer, transform):
        """
        """
        if file_id in self.active_files:
            raise ValueError('file_id=%r already registered for transfer' % file_id)
        deferred = defer.Deferred()
        self.active_files[file_id] = (deferred, file_object, writer, transform)
        if _Debug:
            lg.args(_DebugLevel*2, file_id, file_object, [fid for fid in self.active_files.keys()])
        if not self.consumer.producerPaused:
            self.resumeProducing()
        return deferred

    def stopFileTransfer(self, file_id, reason='cancelled'):
        if file_id not in self.active_files:
            raise ValueError('file_id=%r is not registered for transfer' % file_id)
        deferred, _, _, _ = self.active_files.pop(file_id, (
            None,
            None,
            None,
            None,
        ))
        if _Debug:
            lg.args(_DebugLevel*2, file_id, [fid for fid in self.active_files.keys()])
        if deferred:
            deferred.errback(Exception(reason))

    def resumeProducing(self):
        if _Debug:
            lg.args(_DebugLevel*2, [fid for fid in self.active_files.keys()])
        files_to_be_removed = []
        for file_id in self.active_files.keys():
            deferred, file_object, writer, transform = self.active_files.get(file_id, (
                None,
                None,
                None,
                None,
            ))
            if not file_object:
                lg.warn('did not found file object for file_id=%r' % file_id)
                files_to_be_removed.append(file_id)
                continue
            chunk = b''
            err = False
            if file_object:
                try:
                    chunk = file_object.read(self.CHUNK_SIZE)
                except Exception as exc:
                    lg.exc()
                    chunk = None
                    err = exc
            if not chunk:
                files_to_be_removed.append(file_id)
                if deferred:
                    if err:
                        deferred.errback(err)
                    else:
                        deferred.callback(True)
                continue
            chunk = transform(chunk)
            writer(chunk)
        for file_id in files_to_be_removed:
            self.active_files.pop(file_id)

    def pauseProducing(self):
        if _Debug:
            lg.args(_DebugLevel*2, [fid for fid in self.active_files.keys()])

    def stopProducing(self):
        if _Debug:
            lg.args(_DebugLevel*2, [fid for fid in self.active_files.keys()])



"""
Apache2 config

<VirtualHost *:80>
        ServerName blockchain.bitdust.io
        ServerAlias www.blockchain.bitdust.io
        Redirect / https://blockchain.bitdust.io/
        RewriteEngine on
        RewriteCond %{SERVER_NAME} =blockchain.bitdust.io [OR]
        RewriteCond %{SERVER_NAME} =www.blockchain.bitdust.io
        RewriteRule ^ https://%{SERVER_NAME}%{REQUEST_URI} [END,NE,R=permanent]
</VirtualHost>

<VirtualHost *:443>
        ServerName blockchain.bitdust.io
        ServerAlias www.blockchain.bitdust.io
        ServerAdmin bitdust.io@gmail.com
        DocumentRoot /var/www
        SSLEngine on
        RewriteEngine on
        RewriteRule ^/([a-zA-Z0-9]*)$ http://localhost:19080/$1 [P,L]
        SSLCertificateFile /ssl/domain.cert.pem
        SSLCertificateKeyFile /ssl/private.key.pem
</VirtualHost>
"""

import sys
import time
import sqlite3
import base64

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))
    src_dir_path = _p.dirname(_p.dirname(_p.dirname(_p.abspath(sys.argv[0]))))
    sys.path.insert(0, src_dir_path)
    sys.path.insert(0, _p.join(src_dir_path, 'bitdust_forks', 'Bismuth'))

#------------------------------------------------------------------------------

from twisted.internet import reactor
from twisted.web import server, resource

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import misc

from bitdust.main import settings
from bitdust.main import config

from bitdust.interface import web_html_template

from bitdust.blockchain import bismuth_node

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

_DataDirPath = None
_ExplorerHost = None
_ExplorerPort = None
_WebListener = None

#------------------------------------------------------------------------------


def init():
    global _DataDirPath
    global _ExplorerHost
    global _ExplorerPort
    global _WebListener

    _DataDirPath = settings.ServiceDir('bismuth_blockchain')
    _ExplorerHost = config.conf().getString('services/blockchain-explorer/host', '127.0.0.1')
    _ExplorerPort = config.conf().getInt('services/blockchain-explorer/web-port', 19080)

    root = BlockchainRootPage()
    root.putChild(b'', BlockchainMainPage())
    try:
        _WebListener = reactor.listenTCP(_ExplorerPort, server.Site(root))  # @UndefinedVariable
        if _Debug:
            lg.out(_DebugLevel, '            have started web server at port %d   hostname=%s' % (_ExplorerPort, strng.to_text(_ExplorerHost)))
    except:
        if _Debug:
            lg.err('exception while trying to listen port ' + str(self.web_port))
        lg.exc()
    if _Debug:
        lg.args(_DebugLevel, data_dir_path=_DataDirPath)
    return True


def shutdown():
    global _WebListener
    if _Debug:
        lg.dbg(_DebugLevel, '')
    if _WebListener:
        _WebListener.stopListening()
        if _Debug:
            lg.out(_DebugLevel, '            stopped web listener')
    _WebListener = None
    return True


#------------------------------------------------------------------------------


def execute(cursor, query, param):
    while True:
        try:
            cursor.execute(query, param)
            break
        except Exception as e:
            print('Database query: {} {}'.format(cursor, query))
            print('Database retry reason: {}'.format(e))
            time.sleep(0.2)
    return cursor


#------------------------------------------------------------------------------


class BlockchainMainPage(resource.Resource):

    def render_GET(self, request):
        global _ExplorerHost

        page_size = 500

        page_num = request.args.get(b'page', [])
        if page_num:
            try:
                page_num = int(strng.to_text(page_num[0]))
            except:
                lg.exc()
                page_num = 0
        else:
            page_num = 0

        page_max = page_num + 1

        src = ''

        conn = sqlite3.connect(bismuth_node.nod().ledger_path, timeout=60.0)
        c = conn.cursor()
        execute(c, 'SELECT * FROM transactions ORDER BY block_height DESC, timestamp DESC LIMIT ? OFFSET ?;', (
            page_size,
            page_size*page_num,
        ))
        _all = c.fetchall()
        c.close()
        conn.close()
        conn = None
        c = None

        if len(_all) < page_size:
            page_max = page_num

        view = []
        b = -1
        x_old = 'init'

        for x in _all:
            if x[0] != x_old:
                color_cell = '#F8F8F8'
                if b >= 0:
                    view.append('<tr><td>&nbsp;</td></tr>')  #block separator
            else:
                color_cell = 'white'

            view.append('<tr bgcolor ={}>'.format(color_cell))

            if x[0] != x_old:
                b = b + 1

            txid_formatted = strng.to_text(base64.b64encode(strng.to_bin(x[5]), altchars=b'-_'))
            view.append('<td><a href="/{}">{}</a></td>'.format(
                txid_formatted,
                txid_formatted[:8],
            ))  #TXID

            view.append('<td>{}'.format(time.strftime('%Y/%m/%d %H:%M:%S', time.gmtime(float(x[1])))))
            view.append('<td>{}</td>'.format(x[2][:8]))  #from
            view.append('<td>{}</td>'.format(x[3][:8]))  #to
            view.append('<td>{}</td>'.format(misc.float2str(x[4], mask='%10.12f')))  #amount

            if x_old != x[0]:
                view.append('<td>{}</td>'.format(x[7][:8]))  #block hash
            else:
                view.append('<td>&nbsp;</td>')  #block hash

            view.append('<td>{}</td>'.format(misc.float2str(x[8], mask='%10.12f')))  #fee
            view.append('<td>{}</td>'.format(misc.float2str(x[9], mask='%10.12f')))  #reward

            view.append('<td>{}{}</td>'.format(x[10][:16], '&hellip;' if len(x[10]) > 16 else ''))  #operation
            view.append('<td>{}{}</td>'.format(x[11][:24], '&hellip;' if len(x[11]) > 24 else ''))  #openfield

            if x_old != x[0]:
                view.append('<td>{}</td>'.format(x[0]))  #block height
            else:
                view.append('<td>{}</td>'.format(x[0]))  #block height

            view.append('</tr>')

            x_old = x[0]

        src += '<div id="ui-blockchain-explorer" class="section bg-light">'
        src += '<div class="container">\n'
        src += '<div class="ui-card">\n'
        src += '<div class="card-body">\n'

        src += '<div class="row justify-content-center">\n'
        src += '<h1 align=center>BitDust blockchain explorer</h1>\n'
        src += '</div>\n'

        src += '<div class="row justify-content-center">\n'
        src += '<ul class="pagination pagination-sm">\n'
        src += '<li class="page-item">\n'
        src += '<a class="page-link" href="/?page={}">\n'.format(max(0, page_num - 1))
        src += '<span aria-hidden="true">&laquo;</span>\n'
        src += '</a>\n'
        src += '</li>\n'
        src += '<li class="page-item">\n'
        src += '<a class="page-link" href="/?page={}">\n'.format(min(page_max, page_num + 1))
        src += '<span aria-hidden="true">&raquo;</span>\n'
        src += '</a>\n'
        src += '</li>\n'
        src += '</ul>\n'
        src += '</div>\n'

        src += '<div class="row justify-content-center">\n'

        src += '<style type="text/css">#blockchain-table td {padding: 0px 5px; font-size: 0.8em; font-family: monospace, monospace; }</style>'

        src += '<div class="table-responsive">\n'
        src += '<table id="blockchain-table" class="table table-bordered table-hover">\n'

        src += '<tr bgcolor=white>\n'
        src += '<td><b>TXID</b></td>\n'
        src += '<td><b>Timestamp</b></td>\n'
        src += '<td><b>From</b></td>\n'
        src += '<td><b>To</b></td>\n'
        src += '<td><b>Amount</b></td>\n'
        src += '<td><b>Hash</b></td>\n'
        src += '<td><b>Fee</b></td>\n'
        src += '<td><b>Reward</b></td>\n'
        src += '<td><b>Operation</b></td>\n'
        src += '<td><b>Openfield</b></td>\n'
        src += '<td><b>Block</b></td>\n'
        src += '</tr>\n'

        src += ''.join(view)

        src += '</table>\n'
        src += '</div>\n'
        src += '</div>\n'

        src += '<br>\n'
        src += '<div class="row justify-content-center">\n'
        src += '<ul class="pagination pagination-sm">\n'
        src += '<li class="page-item">\n'
        src += '<a class="page-link" href="/?page={}">\n'.format(max(0, page_num - 1))
        src += '<span aria-hidden="true">&laquo;</span>\n'
        src += '</a>\n'
        src += '</li>\n'
        src += '<li class="page-item">\n'
        src += '<a class="page-link" href="/?page={}">\n'.format(min(page_max, page_num + 1))
        src += '<span aria-hidden="true">&raquo;</span>\n'
        src += '</a>\n'
        src += '</li>\n'
        src += '</ul>\n'
        src += '</div>\n'

        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'

        html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
            title='BitDust blockchain explorer',
            site_url='https://bitdust.io',
            basepath='https://%s/' % _ExplorerHost,
            wikipath='https://bitdust.io/wiki/',
            idserverspath='https://identities.bitdust.io/',
            blockchainpath='https://blockchain.bitdust.io/',
            div_main_class='main blockchain',
            div_main_body=src,
            google_analytics='',
            pre_footer='',
        )
        return strng.to_bin(html_src)


#------------------------------------------------------------------------------


class BlockchainTransactionPage(resource.Resource):

    def __init__(self, tx_id):
        resource.Resource.__init__(self)
        self.tx_id = tx_id

    def render_GET(self, request):
        src = ''
        src += '<div id="ui-blockchain-transaction" class="section bg-light">'
        src += '<div class="container">\n'
        src += '<div class="ui-card">\n'
        src += '<div class="card-body">\n'

        try:
            _t = strng.to_text(base64.b64decode(self.tx_id, altchars='-_'))
        except:
            src += '<p align=center>invalid transaction ID</p>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
                title='BitDust blockchain explorer',
                site_url='https://bitdust.io',
                basepath='https://%s/' % _ExplorerHost,
                wikipath='https://bitdust.io/wiki/',
                idserverspath='https://identities.bitdust.io/',
                blockchainpath='https://blockchain.bitdust.io/',
                div_main_class='main blockchain-transaction',
                div_main_body=src,
                google_analytics='',
                pre_footer='',
            )
            return strng.to_bin(html_src)

        try:
            conn = sqlite3.connect(bismuth_node.nod().ledger_path, timeout=60.0)
            c = conn.cursor()
            execute(c, 'SELECT * FROM transactions WHERE substr(signature,1,4)=substr(?1,1,4) and signature like ?1;', (_t + '%', ))
            raw = c.fetchone()
            c.close()
            conn.close()
            conn = None
            c = None
        except:
            src += '<p align=center>reading failed</p>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
                title='BitDust blockchain explorer',
                site_url='https://bitdust.io',
                basepath='https://%s/' % _ExplorerHost,
                wikipath='https://bitdust.io/wiki/',
                idserverspath='https://identities.bitdust.io/',
                blockchainpath='https://blockchain.bitdust.io/',
                div_main_class='main blockchain-transaction',
                div_main_body=src,
                google_analytics='',
                pre_footer='',
            )
            return strng.to_bin(html_src)

        if not raw:
            src += '<p align=center>invalid transaction ID</p>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
            html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
                title='BitDust blockchain explorer',
                site_url='https://bitdust.io',
                basepath='https://%s/' % _ExplorerHost,
                wikipath='https://bitdust.io/wiki/',
                idserverspath='https://identities.bitdust.io/',
                blockchainpath='https://blockchain.bitdust.io/',
                div_main_class='main blockchain-transaction',
                div_main_body=src,
                google_analytics='',
                pre_footer='',
            )
            return strng.to_bin(html_src)

        src += '<div>block: <b>{}</b></div><br>\n'.format(raw[0])
        src += '<div>block hash: <b><code>{}</code></b></div><br>\n'.format(raw[7])
        src += '<div>timestamp: <b>{}</b></div><br>\n'.format(time.strftime('%Y/%m/%d %H:%M:%S', time.gmtime(float(raw[1]))))
        src += '<div>sender: <b><code>{}</code></b></div><br>\n'.format(raw[2])
        src += '<div>recipient: <b><code>{}</code></b></div><br>\n'.format(raw[3])
        src += '<div>amount: <b>{}</b></div><br>\n'.format(misc.float2str(raw[4], mask='%10.12f'))
        src += '<div>fee: <b>{}</b></div><br>\n'.format(misc.float2str(raw[8], mask='%10.12f'))
        src += '<div>reward: <b>{}</b></div><br>\n'.format(misc.float2str(raw[9], mask='%10.12f'))
        src += '<div>operation: <b>{}</b></div><br>\n'.format(raw[10])
        src += '<div style="margin: 0 auto; overflow-wrap: break-word; word-wrap: break-word;">openfield: <b><code>{}</code></b></div><br>\n'.format(raw[11])
        src += '<div style="margin: 0 auto; overflow-wrap: break-word; word-wrap: break-word;">signature:\n<b><code>{}</code></b></div><br>\n'.format(self.tx_id)
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'

        html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
            title='BitDust blockchain explorer',
            site_url='https://bitdust.io',
            basepath='https://%s/' % _ExplorerHost,
            wikipath='https://bitdust.io/wiki/',
            idserverspath='https://identities.bitdust.io/',
            blockchainpath='https://blockchain.bitdust.io/',
            div_main_class='main blockchain-transaction',
            div_main_body=src,
            google_analytics='',
            pre_footer='',
        )
        return strng.to_bin(html_src)


#------------------------------------------------------------------------------


class BlockchainRootPage(resource.Resource):

    def getChild(self, path, request):
        if not path:
            return self
        try:
            path = strng.to_text(path)
        except:
            return resource.NoResource('Not found')
        if path:
            return BlockchainTransactionPage(path)
        return resource.NoResource('Not found')


#------------------------------------------------------------------------------


def main():
    bismuth_node.init()
    settings.init()
    reactor.addSystemEventTrigger('before', 'shutdown', shutdown)  # @UndefinedVariable
    reactor.callWhenRunning(init)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# network_transport.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (network_transport.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: network_transport.

.. role:: red
BitDust network_transport() Automat


EVENTS:
    * :red:`failed`
    * :red:`init`
    * :red:`receiving-started`
    * :red:`restart`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`stopped`
    * :red:`transport-initialized`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import platform

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import misc
from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.userid import my_id

from bitdust.main import settings

from bitdust.transport import gateway

#------------------------------------------------------------------------------


class NetworkTransport(automat.Automat):
    """
    This class implements all the functionality of the ``network_transport()``
    state machine.
    """

    fast = True

    def __init__(self, proto, interface, state_changed_callback=None):
        self.proto = strng.to_text(proto)
        self.host = None
        self.interface = interface
        self.state_changed_callback = None
        self.options = {}
        automat.Automat.__init__(
            self,
            name='%s_transport' % proto,
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )

    def call(self, method_name, *args):
        method = getattr(self.interface, method_name, None)
        if method is None:
            raise Exception('method %s not found in %r transport' % (method_name, self.proto))
            # lg.err('method %s not found in %r transport' % (method_name, self.proto))
            # return fail(Exception('Method %s not found in the transport %s interface' % (method_name, self.proto)))
        return method(*args)

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when automat's state was changed.
        """
        if self.state_changed_callback:
            self.state_changed_callback(self, oldstate, newstate)
        gateway.on_transport_state_changed(self, oldstate, newstate)

    def state_not_changed(self, curstate, event_string, *args, **kwargs):
        """
        A small hack to catch all events after "verify" processing.
        """
        if self.state_changed_callback:
            self.state_changed_callback(self, curstate, curstate)
        gateway.on_transport_state_changed(self, curstate, curstate)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'INIT'
                self.StartNow = False
                self.StopNow = False
                self.doInit(*args, **kwargs)
        #---STARTING---
        elif self.state == 'STARTING':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'failed':
                self.state = 'OFFLINE'
            elif event == 'receiving-started' and not self.StopNow:
                self.state = 'LISTENING'
                self.doSaveOptions(*args, **kwargs)
            elif event == 'stop':
                self.StopNow = True
            elif event == 'receiving-started' and self.StopNow:
                self.state = 'STOPPING'
                self.StopNow = False
                self.doStop(*args, **kwargs)
            elif event == 'restart':
                self.StopNow = True
                self.StartNow = True
        #---LISTENING---
        elif self.state == 'LISTENING':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doStop(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'STOPPING'
                self.StopNow = False
                self.doStop(*args, **kwargs)
            elif event == 'restart':
                self.state = 'STOPPING'
                self.StopNow = False
                self.StartNow = True
                self.doStop(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start' or event == 'restart':
                self.state = 'STARTING'
                self.StopNow = False
                self.StartNow = False
                self.doStart(*args, **kwargs)
        #---STOPPING---
        elif self.state == 'STOPPING':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stopped' and not self.StartNow:
                self.state = 'OFFLINE'
            elif event == 'stopped' and self.StartNow:
                self.state = 'STARTING'
                self.StartNow = False
                self.doStart(*args, **kwargs)
            elif event == 'start' or event == 'restart':
                self.StartNow = True
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---INIT---
        elif self.state == 'INIT':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'transport-initialized' and self.StartNow:
                self.state = 'STARTING'
                self.doCreateProxy(*args, **kwargs)
                self.StartNow = False
                self.doStart(*args, **kwargs)
            elif event == 'transport-initialized' and not self.StartNow:
                self.state = 'OFFLINE'
                self.doCreateProxy(*args, **kwargs)
            elif event == 'start' or event == 'restart':
                self.StartNow = True
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(8, 'network_transport.doInit : %s' % str(*args, **kwargs))
        gateway.attach(self)
        try:
            listener, state_changed_callback = args[0]
        except:
            listener, state_changed_callback = args[0], None
        self.state_changed_callback = state_changed_callback
        self.interface.init(listener)

    def doStart(self, *args, **kwargs):
        """
        Action method.
        """
        options = {
            'idurl': my_id.getIDURL(),
        }
        id_contact = ''
        default_host = ''
        if self.proto == 'tcp':
            if not id_contact:
                default_host = strng.to_bin(misc.readExternalIP()) + b':' + strng.to_bin(str(settings.getTCPPort()))
            options['host'] = id_contact or default_host
            options['tcp_port'] = settings.getTCPPort()
        elif self.proto == 'udp':
            if not id_contact:
                default_host = strng.to_bin(nameurl.GetName(my_id.getIDURL())) + b'@' + strng.to_bin(platform.node())
            options['host'] = id_contact or default_host
            options['dht_port'] = settings.getDHTPort()
            options['udp_port'] = settings.getUDPPort()
        elif self.proto == 'proxy':
            pass
        elif self.proto == 'http':
            if not id_contact:
                default_host = strng.to_bin(misc.readExternalIP()) + b':' + strng.to_bin(str(settings.getHTTPPort()))
            options['host'] = id_contact or default_host
            options['http_port'] = settings.getHTTPPort()
        if _Debug:
            lg.out(8, 'network_transport.doStart connecting %s transport : %s' % (self.proto.upper(), options))
        self.interface.connect(options)

    def doStop(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(8, 'network_transport.doStop disconnecting %s transport' % (self.proto.upper()))
        self.interface.disconnect()

    def doCreateProxy(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            self.interface.create_proxy(args[0])

    def doSaveOptions(self, *args, **kwargs):
        """
        Action method.
        """
        proto, self.host, self.options = args[0]
        proto = strng.to_text(proto)
        if proto != self.proto:
            lg.warn('wrong protocol: %r' % proto)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        # gateway.transports().pop(self.proto)
        self.interface.shutdown()
        self.destroy()
        self.interface = None
        gateway.detach(self)



#!/usr/bin/python
# coins_index.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (coins_index.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: coins_index
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import six

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

from hashlib import md5

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

#------------------------------------------------------------------------------

if six.PY2:
    from bitdust_forks.CodernityDB.hash_index import HashIndex
    from bitdust_forks.CodernityDB.tree_index import TreeBasedIndex
else:
    from bitdust_forks.CodernityDB3.hash_index import HashIndex
    from bitdust_forks.CodernityDB3.tree_index import TreeBasedIndex

#------------------------------------------------------------------------------


def definitions():
    return [
        (
            'creator',
            Creator,
        ),
        (
            'signer',
            Signer,
        ),
        (
            'miner',
            Miner,
        ),
        (
            'hash',
            HashID,
        ),
        (
            'prev',
            PrevHashID,
        ),
        (
            'time_created',
            TimeCreated,
        ),
        (
            'time_signed',
            TimeSigned,
        ),
        (
            'time_mined',
            TimeMined,
        ),
        (
            'supplier',
            Supplier,
        ),
        (
            'customer',
            Customer,
        ),
        (
            'supplier_customer',
            SupplierCustomer,
        ),
    ]


#------------------------------------------------------------------------------


def make_custom_header():
    src = '\n'
    src += 'from coins.coins_index import BaseHashIndex\n'
    src += 'from coins.coins_index import BaseMD5Index\n'
    src += 'from coins.coins_index import BaseChainIndex\n'
    src += 'from coins.coins_index import BaseTimeIndex\n'
    return src


#------------------------------------------------------------------------------


class BaseHashIndex(HashIndex):
    role = None
    field = None
    key_format = '16s'

    def __init__(self, *args, **kwargs):
        kwargs['key_format'] = self.key_format
        super(BaseHashIndex, self).__init__(*args, **kwargs)

    def transform_key(self, key):
        return key

    def make_key_value(self, data):
        try:
            return self.transform_key(data[self.role][self.field]), None
        except (
            AttributeError,
            ValueError,
            KeyError,
            IndexError,
        ):
            return None
        except Exception:
            lg.exc()

    def make_key(self, key):
        return self.transform_key(key)


#------------------------------------------------------------------------------


class BaseMD5Index(BaseHashIndex):
    def transform_key(self, key):
        return md5(key).digest()


#------------------------------------------------------------------------------


class BaseChainIndex(HashIndex):
    producer_role = None
    consumer_role = None

    def __init__(self, *args, **kwargs):
        kwargs['key_format'] = '16s'
        super(BaseChainIndex, self).__init__(*args, **kwargs)

    def make_key_value(self, data):
        try:
            chain_id = '%s_%s' % (
                data['payload'][self.producer_role],
                data['payload'][self.consumer_role],
            )
            return md5(chain_id).digest(), None
        except (
            AttributeError,
            ValueError,
            KeyError,
            IndexError,
        ):
            return None
        except Exception:
            lg.exc()

    def make_key(self, key):
        return md5(key).digest()


#------------------------------------------------------------------------------


class BaseTimeIndex(TreeBasedIndex):
    role = None

    def __init__(self, *args, **kwargs):
        kwargs['key_format'] = 'I'
        kwargs['node_capacity'] = 128
        super(BaseTimeIndex, self).__init__(*args, **kwargs)

    def make_key_value(self, data):
        try:
            return data[self.role]['time'], None
        except (
            ValueError,
            KeyError,
            IndexError,
        ):
            return None
        except Exception:
            lg.exc()

    def make_key(self, key):
        return key


#------------------------------------------------------------------------------


class Creator(BaseMD5Index):
    role = 'creator'
    field = 'idurl'


#------------------------------------------------------------------------------


class Signer(BaseMD5Index):
    role = 'signer'
    field = 'idurl'


#------------------------------------------------------------------------------


class Miner(BaseMD5Index):
    role = 'miner'
    field = 'idurl'


#------------------------------------------------------------------------------


class Supplier(BaseMD5Index):
    role = 'payload'
    field = 'supplier'


#------------------------------------------------------------------------------


class Customer(BaseMD5Index):
    role = 'payload'
    field = 'customer'


#------------------------------------------------------------------------------


class HashID(BaseHashIndex):
    role = 'miner'
    field = 'hash'
    key_format = '40s'


#------------------------------------------------------------------------------


class PrevHashID(BaseHashIndex):
    role = 'miner'
    field = 'prev'
    key_format = '40s'


#------------------------------------------------------------------------------


class SupplierCustomer(BaseChainIndex):
    producer_role = 'supplier'
    consumer_role = 'customer'


#------------------------------------------------------------------------------


class TimeCreated(BaseTimeIndex):
    role = 'creator'


#------------------------------------------------------------------------------


class TimeSigned(BaseTimeIndex):
    role = 'signer'


#------------------------------------------------------------------------------


class TimeMined(BaseTimeIndex):
    role = 'miner'



#!/usr/bin/env python
# tcp_connection.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tcp_connection.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: tcp_connection.

.. role:: red

BitDust tcp_connection() Automat

EVENTS:
    * :red:`connection-lost`
    * :red:`connection-made`
    * :red:`data-received`
    * :red:`disconnect`
    * :red:`send-keep-alive`
    * :red:`timer-10sec`
"""
#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

import os
import time

from twisted.protocols import basic  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng
from bitdust.lib import net_misc

#------------------------------------------------------------------------------

FIRST_PRIORITY_SHORT_FILE_SIZE = 64*1024

CMD_HELLO = b'h'
CMD_WAZAP = b'w'
CMD_DATA = b'd'
CMD_OK = b'o'
CMD_ABORT = b'a'
CMD_LIST = [
    CMD_HELLO,
    CMD_WAZAP,
    CMD_DATA,
    CMD_OK,
    CMD_ABORT,
]

#------------------------------------------------------------------------------


class TCPConnection(automat.Automat, basic.Int32StringReceiver):
    SoftwareVersion = b'1'

    timers = {
        'timer-10sec': (10.0, ['CLIENT?', 'SERVER?']),
    }

    def __init__(self):
        self.force_keep_alive = False
        self.stream = None
        self.peer_address = None
        self.peer_external_address = None
        self.peer_idurl = None
        self.total_bytes_received = 0
        self.total_bytes_sent = 0
        self.outboxQueue = []
        self.last_wazap_received = 0

    def is_connected(self):
        return self.state in [
            'CONNECTED',
        ]

    def get_proto(self):
        return 'tcp'

    def get_host(self):
        return '%s:%d' % (strng.to_text(self.peer_address[0]), int(self.peer_address[1]))

    def get_idurl(self):
        return self.peer_idurl

    def connectionMade(self):
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.connectionMade %s' % net_misc.pack_address(self.getTransportAddress()))
        address = self.getAddress()
        name = 'tcp_connection[%s]' % strng.to_text(net_misc.pack_address(address))
        automat.Automat.__init__(
            self,
            name=name,
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )
        self.automat('connection-made')

    def connectionLost(self, reason):
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.connectionLost with %s' % net_misc.pack_address(self.getTransportAddress()))
        self.automat('connection-lost')

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'connection-made' and not self.isOutgoing(*args, **kwargs):
                self.state = 'SERVER?'
                self.doInit(*args, **kwargs)
            elif event == 'connection-made' and self.isOutgoing(*args, **kwargs):
                self.state = 'CLIENT?'
                self.doInit(*args, **kwargs)
                self.doCloseOutgoing(*args, **kwargs)
                self.doSendHello(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'data-received':
                self.doReceiveData(*args, **kwargs)
            elif event == 'connection-lost':
                self.state = 'CLOSED'
                self.doStopInOutFiles(*args, **kwargs)
                self.doCloseStream(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect':
                self.state = 'DISCONNECT'
                self.doStopInOutFiles(*args, **kwargs)
                self.doCloseStream(*args, **kwargs)
                self.doDisconnect(*args, **kwargs)
            elif event == 'send-keep-alive':
                self.doSendWazap(*args, **kwargs)
        #---CLIENT?---
        elif self.state == 'CLIENT?':
            if event == 'connection-lost':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'data-received' and self.isWazap(*args, **kwargs) and self.isSomePendingFiles(*args, **kwargs):
                self.state = 'CONNECTED'
                self.doReadWazap(*args, **kwargs)
                self.doOpenStream(*args, **kwargs)
                self.doStartPendingFiles(*args, **kwargs)
            elif event == 'timer-10sec' or event == 'disconnect' or (event == 'data-received' and not (self.isWazap(*args, **kwargs) and self.isSomePendingFiles(*args, **kwargs))):
                self.state = 'DISCONNECT'
                self.doDisconnect(*args, **kwargs)
        #---SERVER?---
        elif self.state == 'SERVER?':
            if event == 'connection-lost':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'data-received' and self.isHello(*args, **kwargs):
                self.state = 'CONNECTED'
                self.doReadHello(*args, **kwargs)
                self.doSendWazap(*args, **kwargs)
                self.doOpenStream(*args, **kwargs)
                self.doStartPendingFiles(*args, **kwargs)
            elif event == 'timer-10sec' or event == 'disconnect' or (event == 'data-received' and not self.isHello(*args, **kwargs)):
                self.state = 'DISCONNECT'
                self.doDisconnect(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---DISCONNECT---
        elif self.state == 'DISCONNECT':
            if event == 'connection-lost':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        return None

    def isHello(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            command, payload = args[0]
            peeraddress, peeridurl = payload.split(b' ')
            peerip, peerport = peeraddress.split(b':')
            peerport = int(peerport)
            peeraddress = (peerip, peerport)
        except:
            return False
        return command == CMD_HELLO

    def isWazap(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            command, payload = args[0]
        except:
            return False
        return command == CMD_WAZAP

    def isOutgoing(self, *args, **kwargs):
        """
        Condition method.
        """
        from bitdust.transport.tcp import tcp_node
        if self.getConnectionAddress() is not None:
            if self.getConnectionAddress() in list(tcp_node.started_connections().keys()):
                return True
        return False

    def isSomePendingFiles(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.factory.pendingoutboxfiles) > 0

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        self.peer_address = self.getTransportAddress()
        self.peer_external_address = self.peer_address
        self.connected = time.time()
        if self.peer_address not in tcp_node.opened_connections():
            tcp_node.opened_connections()[self.peer_address] = []
        tcp_node.opened_connections()[self.peer_address].append(self)
        tcp_node.increase_connections_counter()
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.doInit with %s, total connections to that address : %d' % (self.peer_address, len(tcp_node.opened_connections()[self.peer_address])))

    def doCloseOutgoing(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        conn = tcp_node.started_connections().pop(self.getConnectionAddress())
        conn.connector = None
        # lg.out(18, 'tcp_connection.doCloseOutgoing    %s closed, %d more started' % (
        #     str(self.peer_address), len(tcp_node.started_connections())))

    def doReadHello(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        try:
            command, payload = args[0]
            peeraddress, peeridurl = payload.split(b' ')
            peerip, peerport = peeraddress.split(b':')
            peerport = int(peerport)
            if not peerip:
                lg.warn('unknown peer IP from Hello packet: %r' % args)
                peerip = self.peer_external_address[0]
            peeraddress = (peerip, peerport)
        except:
            lg.exc()
            return
        # self.peer_external_address = (self.peer_external_address[0], peerport)
        self.peer_external_address = peeraddress
        self.peer_idurl = peeridurl
        if self.peer_address != self.peer_external_address:
            tcp_node.opened_connections()[self.peer_address].remove(self)
            if len(tcp_node.opened_connections()[self.peer_address]) == 0:
                tcp_node.opened_connections().pop(self.peer_address)
            old_address = self.peer_address
            self.peer_address = self.peer_external_address
            if self.peer_address not in tcp_node.opened_connections():
                tcp_node.opened_connections()[self.peer_address] = []
            tcp_node.opened_connections()[self.peer_address].append(self)
            if _Debug:
                lg.out(_DebugLevel, '%s : external peer address changed from %s to %s' % (self, old_address, self.peer_address))
        # lg.out(18, 'tcp_connection.doReadHello from %s' % (self.peer_idurl))

    def doReadWazap(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            command, payload = args[0]
        except:
            return
        self.peer_idurl = payload
        # lg.out(18, 'tcp_connection.doReadWazap from %s' % (self.peer_idurl))

    def doReceiveData(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            command, payload = args[0]
        except:
            return
        if command == CMD_DATA:
            self.stream.data_received(payload)
        elif command == CMD_OK:
            self.stream.ok_received(payload)
        elif command == CMD_ABORT:
            self.stream.abort_received(payload)
        elif command == CMD_WAZAP:
            self.last_wazap_received = time.time()
        else:
            pass

    def doSendHello(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        host = strng.to_bin(tcp_node.my_host() or '127.0.0.1:7771')
        idurl = strng.to_bin(tcp_node.my_idurl() or 'None')
        payload = host + b' ' + idurl
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.doSendHello %r to %s' % (payload, net_misc.pack_address(self.getTransportAddress())))
        self.sendData(CMD_HELLO, payload)

    def doSendWazap(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        payload = strng.to_bin(tcp_node.my_idurl() or 'None')
        self.sendData(CMD_WAZAP, payload)

    def doStartPendingFiles(self, *args, **kwargs):
        """
        Action method.
        """
        for filename, description, result_defer, keep_alive in self.factory.pendingoutboxfiles:
            self.append_outbox_file(filename, description, result_defer, keep_alive)
        self.factory.pendingoutboxfiles = []

    def doStopInOutFiles(self, *args, **kwargs):
        """
        Action method.
        """
        if self.state == 'DISCONNECT':
            self.stream.abort_files('disconnecting')
        else:
            self.stream.abort_files()

    def doOpenStream(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_stream
        self.stream = tcp_stream.TCPFileStream(self)

    def doCloseStream(self, *args, **kwargs):
        """
        Action method.
        """
        self.stream.close()
        del self.stream
        self.stream = None

    def doDisconnect(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.doDisconnect with %s %s' % (str(self.peer_address), self.peer_idurl))
        if self.factory:
            for filename, description, result_defer, keep_alive in self.factory.pendingoutboxfiles:
                result_defer.callback((
                    None,
                    'failed',
                    'disconnected',
                ))
            self.factory.pendingoutboxfiles = []
        try:
            self.transport.stopListening()
        except:
            try:
                self.transport.loseConnection()
            except:
                lg.exc()

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.tcp import tcp_node
        self.destroy()
        if self.peer_address in tcp_node.opened_connections():
            tcp_node.opened_connections()[self.peer_address].remove(self)
            if len(tcp_node.opened_connections()[self.peer_address]) == 0:
                tcp_node.opened_connections().pop(self.peer_address)
            tcp_node.decrease_connections_counter()
        else:
            raise Exception('not found %r in the opened connections' % self.peer_address)
        self.stream = None
        self.peer_address = None
        self.peer_external_address = None
        self.peer_idurl = None
        self.outboxQueue = []

    #------------------------------------------------------------------------------

    def getTransportAddress(self):
        peer = self.transport.getPeer()
        return net_misc.normalize_address((
            peer.host,
            int(peer.port),
        ))

    def getConnectionAddress(self):
        return net_misc.normalize_address(self.factory.connection_address)

    def getAddress(self):
        addr = self.getConnectionAddress()
        if not addr:
            addr = self.getTransportAddress()
        return net_misc.normalize_address(addr)

    def sendData(self, command, payload):
        try:
            data = self.SoftwareVersion + strng.to_bin(command.lower()[0:1]) + strng.to_bin(payload)
            self.sendString(data)
        except:
            lg.exc()
            return False
        # self.automat('data-sent', data)
        return True

    def stringReceived(self, data):
        try:
            version = data[0:1]
            command = data[1:2]
            payload = data[2:]
            if version != self.SoftwareVersion:
                raise Exception('different software version')
            if command not in CMD_LIST:
                raise Exception('unknown command received')
        except:
            lg.warn('invalid string received in tcp connection: %r' % data)
            try:
                self.transport.stopListening()
            except:
                try:
                    self.transport.loseConnection()
                except:
                    lg.exc()
            return
        self.automat('data-received', (command, payload))

    def append_outbox_file(self, filename, description='', result_defer=None, keep_alive=True):
        self.outboxQueue.append((filename, description, result_defer, keep_alive))

    def process_outbox_queue(self):
        if self.state != 'CONNECTED':
            return False
        if self.stream is None:
            return False
        from bitdust.transport.tcp import tcp_stream
        has_reads = False
        while len(self.outboxQueue) > 0 and len(self.stream.outboxFiles) < tcp_stream.MAX_SIMULTANEOUS_OUTGOING_FILES:
            filename, description, result_defer, keep_alive = self.outboxQueue.pop(0)
            has_reads = True
            # we have a queue of files to be sent
            # somehow file may be removed before we start sending it
            # so we check it here and skip not existed files
            if not os.path.isfile(filename):
                self.failed_outbox_queue_item(filename, description, 'file not exist')
                if not (keep_alive or self.force_keep_alive):
                    self.automat('shutdown')
                continue
            try:
                filesize = os.path.getsize(filename)
            except:
                self.failed_outbox_queue_item(filename, description, 'can not get file size')
                if not (keep_alive or self.force_keep_alive):
                    self.automat('shutdown')
                continue
            self.stream.create_outbox_file(filename, filesize, description, result_defer, keep_alive)
        return has_reads

    def failed_outbox_queue_item(self, filename, description='', error_message=''):
        from bitdust.transport.tcp import tcp_interface
        if _Debug:
            lg.out(_DebugLevel, 'tcp_connection.failed_outbox_queue_item %s because %s' % (filename, error_message))
        try:
            tcp_interface.interface_cancelled_file_sending(self.getAddress(), filename, 0, description, error_message).addErrback(lambda err: lg.exc(err))
        except Exception as exc:
            lg.warn(str(exc))
